#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import LinuxBridge as LB

from libnmstate.ifaces.linux_bridge import LinuxBridgeIface
from libnmstate.ifaces.ifaces import Ifaces

from ..testlib.bridgelib import LINUX_BRIDGE_IFACE_NAME
from ..testlib.bridgelib import TEST_PORT_NAMES
from ..testlib.bridgelib import PORT1_IFACE_NAME
from ..testlib.bridgelib import PORT1_PORT_CONFIG
from ..testlib.bridgelib import PORT2_IFACE_NAME
from ..testlib.bridgelib import PORT2_PORT_CONFIG
from ..testlib.bridgelib import PORT2_VLAN_CONFIG_TRUNK
from ..testlib.bridgelib import TRUNK_TAGS_ID_RANGES
from ..testlib.bridgelib import gen_bridge_iface_info
from ..testlib.bridgelib import gen_bridge_iface_info_with_vlan_filter
from ..testlib.ifacelib import gen_foo_iface_info

Port = LB.Port
Vlan = LB.Port.Vlan


class TestLinuxBridgeIface:
    def test_linux_bridge_sort_port(self):
        iface_info1 = gen_bridge_iface_info()
        iface_info2 = gen_bridge_iface_info()
        iface_info2[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE].reverse()

        assert (
            LinuxBridgeIface(iface_info1).state_for_verify()
            == LinuxBridgeIface(iface_info2).state_for_verify()
        )

    def test_is_controller(self):
        iface = LinuxBridgeIface(gen_bridge_iface_info())

        assert iface.is_controller

    def test_is_virtual(self):
        iface = LinuxBridgeIface(gen_bridge_iface_info())

        assert iface.is_virtual

    def test_get_port(self):
        iface = LinuxBridgeIface(gen_bridge_iface_info())

        assert iface.port == TEST_PORT_NAMES

    def test_desire_port_name_full_merge_from_current(self):
        cur_iface_info = gen_bridge_iface_info()
        des_iface_info = gen_bridge_iface_info()
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            {Port.NAME: PORT2_IFACE_NAME},
            {Port.NAME: PORT1_IFACE_NAME},
        ]
        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)

        des_iface.merge(cur_iface)
        des_iface.sort_port()
        cur_iface.sort_port()

        assert des_iface.to_dict() == gen_bridge_iface_info()

    def test_desire_port_name_only_merge_config_from_current(self):
        cur_iface_info = gen_bridge_iface_info()
        des_iface_info = gen_bridge_iface_info()
        expected_iface_info = gen_bridge_iface_info()
        expected_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            deepcopy(PORT2_PORT_CONFIG)
        ]
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            {Port.NAME: PORT2_IFACE_NAME},
        ]

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)

        assert des_iface.to_dict() == expected_iface_info

    def test_merged_with_desire_remove_all_ports(self):
        cur_iface_info = gen_bridge_iface_info()
        des_iface_info = gen_bridge_iface_info()
        expected_iface_info = gen_bridge_iface_info()
        expected_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = []
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = []

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)

        assert des_iface.to_dict() == expected_iface_info

    def test_merge_with_desired_ports_not_defined(self):
        cur_iface_info = gen_bridge_iface_info()
        des_iface_info = gen_bridge_iface_info()
        expected_iface_info = gen_bridge_iface_info()
        des_iface_info[LB.CONFIG_SUBTREE].pop(LB.PORT_SUBTREE)

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)

        assert des_iface.to_dict() == expected_iface_info

    def test_desire_port_name_partial_merge_from_current_with_vlan_filter(
        self,
    ):
        cur_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info = gen_bridge_iface_info_with_vlan_filter()
        expected_port_config = deepcopy(PORT2_PORT_CONFIG)
        expected_port_config[Port.VLAN_SUBTREE] = PORT2_VLAN_CONFIG_TRUNK
        expected_iface_info = gen_bridge_iface_info()
        expected_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            expected_port_config
        ]
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            {Port.NAME: PORT2_IFACE_NAME},
        ]

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)

        assert des_iface.to_dict() == expected_iface_info

    def test_fix_access_mode_with_enable_native(self):
        cur_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info = gen_bridge_iface_info_with_vlan_filter()
        expected_iface_info = gen_bridge_iface_info_with_vlan_filter()
        cur_vlan_conf = cur_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0][
            Port.VLAN_SUBTREE
        ]
        cur_vlan_conf[Vlan.ENABLE_NATIVE] = True
        cur_vlan_conf[Vlan.MODE] = Vlan.Mode.TRUNK

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)
        des_iface.pre_edit_validation_and_cleanup()

        assert des_iface.to_dict() == expected_iface_info

    def test_merge_fix_tag_with_trunk_disabled_native(self):
        cur_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0][
            Port.VLAN_SUBTREE
        ] = {
            Vlan.ENABLE_NATIVE: False,
            Vlan.MODE: Vlan.Mode.TRUNK,
            Vlan.TRUNK_TAGS: [{Vlan.TrunkTags.ID: 101}],
        }
        expected_iface_info = deepcopy(des_iface_info)

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)
        des_iface.pre_edit_validation_and_cleanup()

        assert des_iface.to_dict() == expected_iface_info

    def test_remove_trunk_tag_if_access_mode(self):
        cur_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info = gen_bridge_iface_info_with_vlan_filter()
        des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1][
            Port.VLAN_SUBTREE
        ] = {
            Vlan.MODE: Vlan.Mode.ACCESS,
        }
        expected_iface_info = gen_bridge_iface_info_with_vlan_filter()
        expected_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1][
            Port.VLAN_SUBTREE
        ] = {Vlan.MODE: Vlan.Mode.ACCESS, Vlan.TAG: 105}

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)
        des_iface.merge(cur_iface)
        des_iface.pre_edit_validation_and_cleanup()

        assert des_iface.to_dict() == expected_iface_info

    def test_gen_metadata_save_port_config_to_port_iface(self):
        br_iface_info = gen_bridge_iface_info()
        port1_iface_info = gen_foo_iface_info()
        port1_iface_info[Interface.NAME] = PORT1_IFACE_NAME
        port2_iface_info = gen_foo_iface_info()
        port2_iface_info[Interface.NAME] = PORT2_IFACE_NAME
        ifaces = Ifaces(
            des_iface_infos=[
                br_iface_info,
                port1_iface_info,
                port2_iface_info,
            ],
            cur_iface_infos=[],
        )
        br_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]
        br_iface.gen_metadata(ifaces)
        br_iface.pre_edit_validation_and_cleanup()
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]

        assert port1_iface.controller == LINUX_BRIDGE_IFACE_NAME
        assert port2_iface.controller == LINUX_BRIDGE_IFACE_NAME
        assert (
            port1_iface.to_dict()[LinuxBridgeIface.BRPORT_OPTIONS_METADATA]
            == PORT1_PORT_CONFIG
        )
        assert (
            port2_iface.to_dict()[LinuxBridgeIface.BRPORT_OPTIONS_METADATA]
            == PORT2_PORT_CONFIG
        )

    def test_gen_metadata_skip_on_absent_iface(self):
        br_iface_info = gen_bridge_iface_info()
        br_iface_info[Interface.STATE] = InterfaceState.ABSENT
        port1_iface_info = gen_foo_iface_info()
        port1_iface_info[Interface.NAME] = PORT1_IFACE_NAME
        port2_iface_info = gen_foo_iface_info()
        port2_iface_info[Interface.NAME] = PORT2_IFACE_NAME
        ifaces = Ifaces(
            des_iface_infos=[
                br_iface_info,
                port1_iface_info,
                port2_iface_info,
            ],
            cur_iface_infos=[],
        )
        br_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]
        br_iface.gen_metadata(ifaces)
        br_iface.pre_edit_validation_and_cleanup()
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]
        assert port1_iface.controller is None
        assert port2_iface.controller is None
        assert (
            LinuxBridgeIface.BRPORT_OPTIONS_METADATA
            not in port1_iface.to_dict()
        )
        assert (
            LinuxBridgeIface.BRPORT_OPTIONS_METADATA
            not in port2_iface.to_dict()
        )

    def test_state_for_verify_normalize_port_vlan(self):
        iface = LinuxBridgeIface(gen_bridge_iface_info())
        expected_iface_info = gen_bridge_iface_info()
        for port_config in expected_iface_info[LB.CONFIG_SUBTREE][
            LB.PORT_SUBTREE
        ]:
            port_config[Port.VLAN_SUBTREE] = {}

        assert iface.state_for_verify() == expected_iface_info

    def test_validate_access_port_with_trunk_tag(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_config = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        port_config[Port.VLAN_SUBTREE][Vlan.TRUNK_TAGS] = [
            {Vlan.TrunkTags.ID: 101}
        ]

        iface = LinuxBridgeIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_trunk_port_without_trunk_tag(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_config = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1]
        port_config[Port.VLAN_SUBTREE].pop(Vlan.TRUNK_TAGS)

        iface = LinuxBridgeIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_trunk_mixing_id_and_id_range(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_config = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1]
        port_config[Port.VLAN_SUBTREE][Vlan.TRUNK_TAGS][0].update(
            TRUNK_TAGS_ID_RANGES[0]
        )

        iface = LinuxBridgeIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_trunk_id_range_missing_min_or_max(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_config = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1]
        port_config[Port.VLAN_SUBTREE][Vlan.TRUNK_TAGS][0] = {
            Vlan.TrunkTags.ID_RANGE: {Vlan.TrunkTags.MIN_RANGE: 400}
        }

        iface = LinuxBridgeIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

        port_config[Port.VLAN_SUBTREE][Vlan.TRUNK_TAGS][0] = {
            Vlan.TrunkTags.ID_RANGE: {Vlan.TrunkTags.MAX_RANGE: 400}
        }

        iface = LinuxBridgeIface(iface_info)
        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_tag_in_access_mode_and_native_trunk(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        iface = LinuxBridgeIface(iface_info)

        iface.pre_edit_validation_and_cleanup()

    def test_validate_tag_in_non_native_trunk(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_conf = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][1]
        port_conf[Port.VLAN_SUBTREE][Vlan.ENABLE_NATIVE] = False
        iface = LinuxBridgeIface(iface_info)

        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_validate_enable_native_in_access_mode(self):
        iface_info = gen_bridge_iface_info_with_vlan_filter()
        port_conf = iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        port_conf[Port.VLAN_SUBTREE][Vlan.ENABLE_NATIVE] = True
        iface = LinuxBridgeIface(iface_info)

        with pytest.raises(NmstateValueError):
            iface.pre_edit_validation_and_cleanup()

    def test_remove_port(self):
        iface_info = gen_bridge_iface_info()
        expected_iface_info = gen_bridge_iface_info()
        expected_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE] = [
            deepcopy(PORT1_PORT_CONFIG)
        ]

        iface = LinuxBridgeIface(iface_info)
        iface.remove_port(PORT2_IFACE_NAME)

        assert iface.to_dict() == expected_iface_info

    def test_verify_when_vlantree_not_defined_in_desire(self):
        iface_info = gen_bridge_iface_info()
        expected_iface_info = gen_bridge_iface_info()
        for port_config in expected_iface_info[LB.CONFIG_SUBTREE][
            LB.PORT_SUBTREE
        ]:
            port_config[Port.VLAN_SUBTREE] = {}

        iface = LinuxBridgeIface(iface_info)

        assert iface.state_for_verify() == expected_iface_info

    def test_get_config_changed_port(self):
        des_iface_info = gen_bridge_iface_info()
        des_port_config = des_iface_info[LB.CONFIG_SUBTREE][LB.PORT_SUBTREE][0]
        des_port_config[Port.STP_PATH_COST] += 1
        cur_iface_info = gen_bridge_iface_info()

        des_iface = LinuxBridgeIface(des_iface_info)
        cur_iface = LinuxBridgeIface(cur_iface_info)

        assert des_iface.config_changed_port(cur_iface) == [PORT1_IFACE_NAME]

    def test_skip_metadata_generation_when_is_absent(self):
        br_iface_info = gen_bridge_iface_info()
        br_iface_info[Interface.STATE] = InterfaceState.ABSENT
        port1_iface_info = gen_foo_iface_info()
        port1_iface_info[Interface.NAME] = PORT1_IFACE_NAME
        port2_iface_info = gen_foo_iface_info()
        port2_iface_info[Interface.NAME] = PORT2_IFACE_NAME
        ifaces = Ifaces(
            des_iface_infos=[
                br_iface_info,
                port1_iface_info,
                port2_iface_info,
            ],
            cur_iface_infos=[],
        )
        br_iface = ifaces.all_kernel_ifaces[LINUX_BRIDGE_IFACE_NAME]
        br_iface.gen_metadata(ifaces)
        br_iface.pre_edit_validation_and_cleanup()
        port1_iface = ifaces.all_kernel_ifaces[PORT1_IFACE_NAME]
        port2_iface = ifaces.all_kernel_ifaces[PORT2_IFACE_NAME]

        assert port1_iface.controller is None
        assert port2_iface.controller is None
        assert (
            LinuxBridgeIface.BRPORT_OPTIONS_METADATA
            not in port1_iface.to_dict()
        )
        assert (
            LinuxBridgeIface.BRPORT_OPTIONS_METADATA
            not in port2_iface.to_dict()
        )



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from contextlib import contextmanager
import copy
import re

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge
from libnmstate.schema import OVSInterface

from . import cmdlib


class Bridge:
    def __init__(self, name):
        self._name = name
        self._ifaces = [
            {
                Interface.NAME: name,
                Interface.TYPE: InterfaceType.OVS_BRIDGE,
                OVSBridge.CONFIG_SUBTREE: {},
            }
        ]
        self._bridge_iface = self._ifaces[0]

    def set_options(self, options):
        self._bridge_iface[OVSBridge.CONFIG_SUBTREE][
            OVSBridge.OPTIONS_SUBTREE
        ] = options

    def set_ovs_db(self, ovs_db_config):
        self._bridge_iface[OVSBridge.OVS_DB_SUBTREE] = ovs_db_config

    def add_system_port(self, name):
        self._add_port(name)

    def add_link_aggregation_port(self, name, port, mode=None):
        self._add_port(name)
        port_cfg = self._get_port(name)
        port_cfg[OVSBridge.Port.LINK_AGGREGATION_SUBTREE] = {
            OVSBridge.Port.LinkAggregation.PORT_SUBTREE: [
                {OVSBridge.Port.LinkAggregation.Port.NAME: port_member}
                for port_member in port
            ]
        }
        if mode:
            port[OVSBridge.Port.LINK_AGGREGATION_SUBTREE][
                OVSBridge.Port.LinkAggregation.MODE
            ] = mode

    def add_internal_port(
        self,
        name,
        *,
        mac=None,
        ipv4_state=None,
        ovs_db=None,
        patch_state=None,
        mtu=None,
    ):
        ifstate = {
            Interface.NAME: name,
            Interface.TYPE: InterfaceType.OVS_INTERFACE,
        }
        if mac:
            ifstate[Interface.MAC] = mac
        if mtu:
            ifstate[Interface.MTU] = mtu
        if ipv4_state:
            ifstate[Interface.IPV4] = ipv4_state
        if ovs_db:
            ifstate[OVSInterface.OVS_DB_SUBTREE] = ovs_db
        if patch_state:
            ifstate[OVSInterface.PATCH_CONFIG_SUBTREE] = patch_state

        self._add_port(name)
        self._ifaces.append(ifstate)

    @property
    def ports_names(self):
        return [port[OVSBridge.Port.NAME] for port in self._get_ports()]

    def _add_port(self, name):
        self._bridge_iface[OVSBridge.CONFIG_SUBTREE].setdefault(
            OVSBridge.PORT_SUBTREE, []
        ).append({OVSBridge.Port.NAME: name})

    def _get_port(self, name):
        ports = self._get_ports()
        return next(
            (port for port in ports if port[OVSBridge.Port.NAME] == name), None
        )

    def _get_ports(self):
        return self._bridge_iface[OVSBridge.CONFIG_SUBTREE].get(
            OVSBridge.PORT_SUBTREE, []
        )

    def set_port_option(self, port_name, new_option):
        for port_option in self._bridge_iface[OVSBridge.CONFIG_SUBTREE].get(
            OVSBridge.PORT_SUBTREE, []
        ):
            if port_option[OVSBridge.Port.NAME] == port_name:
                port_option.update(new_option)

    @contextmanager
    def create(self):
        desired_state = self.state
        libnmstate.apply(desired_state)
        try:
            yield desired_state
        finally:
            desired_state = {
                Interface.KEY: _set_ifaces_state(
                    self._ifaces, InterfaceState.ABSENT
                )
            }
            libnmstate.apply(desired_state, verify_change=False)

    def apply(self):
        libnmstate.apply(self.state)

    @property
    def state(self):
        return {
            Interface.KEY: _set_ifaces_state(self._ifaces, InterfaceState.UP)
        }


def _set_ifaces_state(ifaces, state):
    ifaces = copy.deepcopy(ifaces)
    for iface in ifaces:
        iface[Interface.STATE] = state
    return ifaces


def get_nm_active_profiles():
    all_profiles_output = cmdlib.exec_cmd(
        "nmcli -g NAME connection show --active".split(" "), check=True
    )[1]
    return all_profiles_output.split("\n")


def get_proxy_port_profile_name_of_ovs_interface(iface_name):
    proxy_port_iface_name = cmdlib.exec_cmd(
        f"nmcli -g connection.master connection show {iface_name}".split(" "),
        check=True,
    )[1].strip()
    all_profiles_output = cmdlib.exec_cmd(
        "nmcli -g NAME,DEVICE connection show".split(" "), check=True
    )[1]
    proxy_port_re = re.compile(
        f"^(?P<profile_name>.+):{proxy_port_iface_name}$"
    )
    for line in all_profiles_output.split("\n"):
        match = proxy_port_re.match(line)
        if match:
            return match.group("profile_name")



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from copy import deepcopy

import pytest

from libnmstate.error import NmstateInternalError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState

from libnmstate.ifaces.base_iface import BaseIface
from ..testlib.constants import MAC_ADDRESS1
from ..testlib.constants import IPV6_LINK_LOCAL_ADDRESS1
from ..testlib.constants import IPV6_ADDRESSES
from ..testlib.ifacelib import gen_foo_iface_info


class TestBaseIface:
    def test_merge(self):
        des_info = gen_foo_iface_info()
        des_info.update(
            {"list_a": [], "dict_b": {"item_a": 1, "dict_c": {"item_b": 2}}}
        )
        cur_info = gen_foo_iface_info()
        cur_info.update(
            {
                "state": "up",
                "list_a": ["a", "b", "c"],
                "dict_b": {
                    "item_a1": 11,
                    "dict_c": {"item_b": 3, "item_d": 4},
                },
            }
        )
        ori_cur_info = deepcopy(cur_info)
        ori_des_info = deepcopy(des_info)
        des_iface = BaseIface(des_info)
        cur_iface = BaseIface(cur_info)
        des_iface.merge(cur_iface)

        expected_info = gen_foo_iface_info()
        expected_info.update(
            {
                "list_a": [],
                "dict_b": {
                    "item_a": 1,
                    "item_a1": 11,
                    "dict_c": {"item_b": 2, "item_d": 4},
                },
            }
        )

        assert cur_info == ori_cur_info
        assert des_info == ori_des_info
        assert cur_iface.to_dict() == ori_cur_info
        assert des_iface.to_dict() == expected_info

    def test_do_not_merge_down_state_from_current(self):
        iface_info = gen_foo_iface_info()
        cur_iface_info = gen_foo_iface_info()
        iface_info.pop(Interface.STATE)
        cur_iface_info[Interface.STATE] = InterfaceState.DOWN

        iface = BaseIface(iface_info)
        cur_iface = BaseIface(cur_iface_info)

        iface.merge(cur_iface)

        assert iface.is_up

    def test_capitalize_mac(self):
        iface_info = gen_foo_iface_info()
        iface_info.update({Interface.MAC: MAC_ADDRESS1})

        expected_info = deepcopy(iface_info)
        iface_info[Interface.MAC] = iface_info[Interface.MAC].lower()

        iface = BaseIface(iface_info)

        assert iface.state_for_verify() == expected_info

    def test_iface_match_other_has_more(self):
        iface = BaseIface(gen_foo_iface_info())
        other_iface_info = gen_foo_iface_info()
        other_iface_info["foo_a"] = "b"
        other_iface = BaseIface(other_iface_info)

        assert iface.match(other_iface)

    def test_iface_match_other_has_less(self):
        iface_info = gen_foo_iface_info()
        iface_info["foo_a"] = "b"
        iface = BaseIface(iface_info)
        other_iface = BaseIface(gen_foo_iface_info())

        assert not iface.match(other_iface)

    def test_iface_match_other_has_diff(self):
        iface_info = gen_foo_iface_info()
        iface_info["foo_a"] = "b"
        iface = BaseIface(iface_info)
        other_iface_info = gen_foo_iface_info()
        other_iface_info["foo_a"] = "c"
        other_iface = BaseIface(other_iface_info)

        assert not iface.match(other_iface)

    def test_state_for_verify_remove_link_local_address(self):
        iface_info = gen_foo_iface_info()
        ipv6_info = {
            InterfaceIPv6.ENABLED: True,
            InterfaceIPv6.ADDRESS: IPV6_ADDRESSES,
        }
        iface_info[Interface.IPV6] = ipv6_info
        expected_iface_info = deepcopy(iface_info)

        iface_info[Interface.IPV6][InterfaceIPv6.ADDRESS].append(
            {
                InterfaceIPv6.ADDRESS_IP: IPV6_LINK_LOCAL_ADDRESS1,
                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
            }
        )
        iface = BaseIface(iface_info)

        assert iface.state_for_verify() == expected_iface_info

    def test_state_for_verify_empty_description(self):
        iface_info = gen_foo_iface_info()
        expected_iface_info = deepcopy(iface_info)

        iface_info[Interface.DESCRIPTION] = ""

        iface = BaseIface(iface_info)

        assert iface.state_for_verify() == expected_iface_info

    def test_state_for_verify_remove_undesired_data(self):
        iface_info = gen_foo_iface_info()
        expected_iface_info = deepcopy(iface_info)

        iface = BaseIface(iface_info)
        iface.raw["foo_a"] = "b"

        assert iface.state_for_verify() == expected_iface_info

    def test_remove_port(self):
        iface = BaseIface(gen_foo_iface_info())

        with pytest.raises(NmstateInternalError):
            iface.remove_port("port_name")

    def test_is_virtual(self):
        iface = BaseIface(gen_foo_iface_info())
        assert iface.is_virtual is False

    def test_create_virtual_port(self):
        iface = BaseIface(gen_foo_iface_info())
        assert iface.create_virtual_port("port_name") is None

    def test_config_changed_port(self):
        iface = BaseIface(gen_foo_iface_info())
        iface2 = BaseIface(gen_foo_iface_info())
        assert iface.config_changed_port(iface2) == []



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import socket

from libnmstate.nm import dns as nm_dns
from libnmstate.nm import route as nm_route
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import Route

from ..ifaces import BaseIface
from .common import NM

INT32_MAX = 2 ** 31 - 1


def create_setting(config, base_con_profile):
    setting_ipv4 = None
    if base_con_profile and config and config.get(InterfaceIPv4.ENABLED):
        setting_ipv4 = base_con_profile.get_setting_ip4_config()
        if setting_ipv4:
            setting_ipv4 = setting_ipv4.duplicate()
            setting_ipv4.clear_addresses()
            setting_ipv4.props.ignore_auto_routes = False
            setting_ipv4.props.never_default = False
            setting_ipv4.props.ignore_auto_dns = False
            setting_ipv4.props.gateway = None
            setting_ipv4.props.route_table = Route.USE_DEFAULT_ROUTE_TABLE
            setting_ipv4.props.route_metric = Route.USE_DEFAULT_METRIC
            setting_ipv4.clear_routes()
            setting_ipv4.clear_routing_rules()
            setting_ipv4.clear_dns()
            setting_ipv4.clear_dns_searches()
            setting_ipv4.props.dns_priority = nm_dns.DEFAULT_DNS_PRIORITY

    if not setting_ipv4:
        setting_ipv4 = NM.SettingIP4Config.new()

    setting_ipv4.props.dhcp_client_id = "mac"
    setting_ipv4.props.method = NM.SETTING_IP4_CONFIG_METHOD_DISABLED
    if config and config.get(InterfaceIPv4.ENABLED):
        if config.get(InterfaceIPv4.DHCP):
            setting_ipv4.props.method = NM.SETTING_IP4_CONFIG_METHOD_AUTO
            setting_ipv4.props.ignore_auto_routes = not config.get(
                InterfaceIPv4.AUTO_ROUTES, True
            )
            setting_ipv4.props.never_default = not config.get(
                InterfaceIPv4.AUTO_GATEWAY, True
            )
            setting_ipv4.props.ignore_auto_dns = not config.get(
                InterfaceIPv4.AUTO_DNS, True
            )
            setting_ipv4.props.route_table = config.get(
                InterfaceIPv4.AUTO_ROUTE_TABLE_ID,
                Route.USE_DEFAULT_ROUTE_TABLE,
            )
            # NetworkManager will remove the virtual interfaces like bridges
            # when the DHCP timeout expired, set it to the maximum value to
            # make this unlikely.
            setting_ipv4.props.dhcp_timeout = INT32_MAX
        elif config.get(InterfaceIPv4.ADDRESS):
            setting_ipv4.props.method = NM.SETTING_IP4_CONFIG_METHOD_MANUAL
            _add_addresses(setting_ipv4, config[InterfaceIPv4.ADDRESS])
        nm_route.add_routes(
            setting_ipv4, config.get(BaseIface.ROUTES_METADATA, [])
        )
        nm_dns.add_dns(setting_ipv4, config.get(BaseIface.DNS_METADATA, {}))
        nm_route.add_route_rules(
            setting_ipv4,
            socket.AF_INET,
            config.get(BaseIface.ROUTE_RULES_METADATA, []),
        )
    return setting_ipv4


def _add_addresses(setting_ipv4, addresses):
    for address in addresses:
        naddr = NM.IPAddress.new(
            socket.AF_INET,
            address[InterfaceIPv4.ADDRESS_IP],
            address[InterfaceIPv4.ADDRESS_PREFIX_LENGTH],
        )
        setting_ipv4.add_address(naddr)


def get_ipv4_config(nm_profile, full_config=False):
    """
    If full_config is False, only provides information regarding:
        * Enable status
        * DHCP status
    """
    if nm_profile is None:
        # Neither unmanaged or not active, let nispor determine its state
        return {}

    info = {InterfaceIPv4.ENABLED: False, InterfaceIPv4.DHCP: False}
    ip_profile = nm_profile.get_setting_ip4_config() if nm_profile else None
    if ip_profile:
        method = ip_profile.get_method()
        if method == NM.SETTING_IP4_CONFIG_METHOD_DISABLED:
            info[InterfaceIPv4.ENABLED] = False
        else:
            info[InterfaceIPv4.ENABLED] = True
            if method == NM.SETTING_IP4_CONFIG_METHOD_AUTO:
                info[InterfaceIPv4.DHCP] = True
                props = ip_profile.props
                info[InterfaceIPv4.AUTO_ROUTES] = not props.ignore_auto_routes
                info[InterfaceIPv4.AUTO_GATEWAY] = not props.never_default
                info[InterfaceIPv4.AUTO_DNS] = not props.ignore_auto_dns
                info[InterfaceIPv4.AUTO_ROUTE_TABLE_ID] = props.route_table
        if full_config:
            info[InterfaceIPv4.ADDRESS] = []
            for nm_addr in ip_profile.props.addresses:
                info[InterfaceIPv4.ADDRESS].append(
                    {
                        InterfaceIPv4.ADDRESS_IP: nm_addr.get_address(),
                        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: int(
                            nm_addr.get_prefix()
                        ),
                    }
                )

    return info


def get_ip_profile(active_connection):
    """
    Get NMSettingIP4Config from NMActiveConnection.
    For any error, return None.
    """
    remote_conn = active_connection.get_connection()
    if remote_conn:
        return remote_conn.get_setting_ip4_config()
    return None


def acs_and_ip_profiles(nm_client):
    for ac in nm_client.get_active_connections():
        ip_profile = get_ip_profile(ac)
        if not ip_profile:
            continue
        yield ac, ip_profile


def is_dynamic(active_connection):
    ip_profile = get_ip_profile(active_connection)
    if ip_profile:
        return ip_profile.get_method() == NM.SETTING_IP4_CONFIG_METHOD_AUTO
    return False



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from contextlib import contextmanager
from copy import deepcopy
import logging
import json
from operator import itemgetter
import os
import time

import pytest

import libnmstate
from libnmstate.schema import Constants
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import Route as RT

from libnmstate.error import NmstateNotImplementedError
from libnmstate.iplib import is_ipv6_link_local_addr

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import bondlib
from .testlib import ifacelib
from .testlib import statelib
from .testlib.ifacelib import get_mac_address
from .testlib.bridgelib import add_port_to_bridge
from .testlib.bridgelib import create_bridge_subtree_state
from .testlib.bridgelib import linux_bridge
from .testlib.retry import retry_till_true_or_timeout
from .testlib.veth import create_veth_pair
from .testlib.veth import remove_veth_pair

ETH1 = "eth1"

DEFAULT_TIMEOUT = 20
NM_DHCP_TIMEOUT_DEFAULT = 45
# The default IPv6 RA/Autoconf timeout is 30 seconds, less than above.
NM_IPV6_AUTOCONF_TIMEOUT_DEFAULT = 30

IPV4_ADDRESS1 = "192.0.2.251"
IPV4_ADDRESS2 = "192.0.2.252"
IPV6_ADDRESS1 = "2001:db8:1::1"
IPV6_ADDRESS2 = "2001:db8:2::1"
IPV6_ADDRESS3 = "2001:db8:1::3"
IPV4_NETWORK1 = "203.0.113.0/24"
IPV6_NETWORK1 = "2001:db8:2::/64"
IPV4_CLASSLESS_ROUTE_DST_NET1 = "198.51.100.0/24"
IPV4_CLASSLESS_ROUTE_NEXT_HOP1 = "192.0.2.1"
IPV6_CLASSLESS_ROUTE_PREFIX = "2001:db8:f"
IPV6_CLASSLESS_ROUTE_DST_NET1 = "{}::/64".format(IPV6_CLASSLESS_ROUTE_PREFIX)

TEST_BRIDGE_NIC = "brtest0"

DHCP_SRV_NIC = "dhcpsrv"
DHCP_CLI_NIC = "dhcpcli"
DHCP_SRV_NS = "nmstate_dhcp_test"
DHCP_SRV_IP4 = IPV4_ADDRESS1
DHCP_SRV_IP6 = IPV6_ADDRESS1
DHCP_SRV_IP6_2 = "{}::1".format(IPV6_CLASSLESS_ROUTE_PREFIX)
DHCP_SRV_IP4_PREFIX = "192.0.2"
DHCP_SRV_IP6_PREFIX = "2001:db8:1"
DHCP_SRV_IP6_NETWORK = "{}::/64".format(DHCP_SRV_IP6_PREFIX)

IPV6_DEFAULT_GATEWAY = "::/0"
IPV4_DEFAULT_GATEWAY = "0.0.0.0/0"

IPV4_DNS_NAMESERVER = "8.8.8.8"
IPV6_DNS_NAMESERVER = "2001:4860:4860::8888"
EXAMPLE_SEARCHES = ["example.org", "example.com"]

DNSMASQ_CONF_STR = """
leasefile-ro
interface={iface}
dhcp-range={ipv4_prefix}.200,{ipv4_prefix}.250,255.255.255.0,48h
enable-ra
dhcp-range={ipv6_prefix}::100,{ipv6_prefix}::fff,ra-names,slaac,64,480h
dhcp-range={ipv6_classless_route}::100,{ipv6_classless_route}::fff,static
dhcp-option=option:classless-static-route,{classless_rt},{classless_rt_dst}
dhcp-option=option:dns-server,{v4_dns_server}
""".format(
    **{
        "iface": DHCP_SRV_NIC,
        "ipv4_prefix": DHCP_SRV_IP4_PREFIX,
        "ipv6_prefix": DHCP_SRV_IP6_PREFIX,
        "classless_rt": IPV4_CLASSLESS_ROUTE_DST_NET1,
        "classless_rt_dst": IPV4_CLASSLESS_ROUTE_NEXT_HOP1,
        "v4_dns_server": DHCP_SRV_IP4,
        "ipv6_classless_route": IPV6_CLASSLESS_ROUTE_PREFIX,
    }
)

DNSMASQ_CONF_PATH = "/etc/dnsmasq.d/nmstate.conf"
# Docker does not allow NetworkManager to edit /etc/resolv.conf.
# Have to read NetworkManager internal resolv.conf
RESOLV_CONF_PATH = "/var/run/NetworkManager/resolv.conf"

parametrize_ip_ver = pytest.mark.parametrize(
    "ip_ver",
    [(Interface.IPV4,), (Interface.IPV6,), (Interface.IPV4, Interface.IPV6)],
    ids=["ipv4", "ipv6", "ipv4&6"],
)


@pytest.fixture(scope="module")
def dhcp_env():
    try:
        create_veth_pair(DHCP_CLI_NIC, DHCP_SRV_NIC, DHCP_SRV_NS)
        _setup_dhcp_nics()

        with open(DNSMASQ_CONF_PATH, "w") as fd:
            fd.write(DNSMASQ_CONF_STR)
        cmdlib.exec_cmd(
            f"ip netns exec {DHCP_SRV_NS} "
            f"dnsmasq -C {DNSMASQ_CONF_PATH}".split(),
            check=True,
        )

        yield
    finally:
        _clean_up()


@pytest.fixture
def dhcpcli_up(dhcp_env):
    with ifacelib.iface_up(DHCP_CLI_NIC) as ifstate:
        yield ifstate


@pytest.fixture
def dhcpcli_up_with_dynamic_ip(dhcp_env):
    with iface_with_dynamic_ip_up(DHCP_CLI_NIC) as ifstate:
        yield ifstate


@contextmanager
def iface_with_dynamic_ip_up(ifname):
    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: ifname,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: _create_ipv4_state(enabled=True, dhcp=True),
                Interface.IPV6: _create_ipv6_state(
                    enabled=True, dhcp=True, autoconf=True
                ),
            }
        ]
    }
    try:
        libnmstate.apply(desired_state)
        assert _poll(_has_ipv4_dhcp_gateway)
        assert _poll(_has_dhcpv4_addr)
        assert _poll(_has_ipv6_auto_gateway)
        assert _poll(_has_dhcpv6_addr)
        yield statelib.show_only((ifname,))
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: ifname,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            }
        )


@pytest.mark.tier1
def test_ipv4_dhcp(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)

    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)


def test_ipv6_dhcp_only(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_nameserver)
    assert _poll(_has_dhcpv6_addr)
    # DHCPv6 does not provide routes
    assert not _has_ipv6_auto_gateway()
    assert not _has_ipv6_auto_extra_route()


@pytest.mark.tier1
def test_ipv6_dhcp_and_autoconf(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_ipv6_auto_extra_route)
    assert _poll(_has_ipv6_auto_nameserver)


def test_dhcp_with_addresses(dhcpcli_up):
    ipv4_state = _create_ipv4_state(enabled=True, dhcp=True)
    ipv4_state[InterfaceIPv4.ADDRESS] = [
        create_ipv4_address_state(IPV4_ADDRESS1, 24),
        create_ipv4_address_state(IPV4_ADDRESS2, 24),
    ]
    ipv6_state = _create_ipv6_state(enabled=True, dhcp=True, autoconf=True)
    ipv6_state[InterfaceIPv6.ADDRESS] = [
        create_ipv6_address_state(IPV6_ADDRESS1, 64),
        create_ipv6_address_state(IPV6_ADDRESS2, 64),
    ]

    desired_state = {
        Interface.KEY: [
            {
                Interface.NAME: DHCP_CLI_NIC,
                Interface.STATE: InterfaceState.UP,
                Interface.IPV4: ipv4_state,
                Interface.IPV6: ipv6_state,
            }
        ]
    }

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)


@pytest.mark.tier1
def test_ipv4_dhcp_on_bond(dhcpcli_up):
    ipv4_state = {Interface.IPV4: _create_ipv4_state(enabled=True, dhcp=True)}
    with bondlib.bond_interface(
        "bond99", port=[DHCP_CLI_NIC], extra_iface_state=ipv4_state
    ) as desired_state:
        assertlib.assert_state_match(desired_state)


def test_ipv4_dhcp_ignore_gateway(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True, auto_gateway=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_classless_route)
    assert not _has_ipv4_dhcp_gateway()


def test_ipv4_dhcp_ignore_dns(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True, auto_dns=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)
    assert not _has_ipv4_dhcp_nameserver()


def test_ipv4_dhcp_ignore_routes(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True, auto_routes=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_nameserver)
    assert not _has_ipv4_dhcp_gateway()
    assert not _has_ipv4_classless_route()


def test_ipv4_dhcp_set_table_id(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True, table_id=100
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)


def test_ipv6_dhcp_set_table_id_without_autoconf(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, table_id=100
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)


def test_ipv6_dhcp_set_table_id_with_autoconf(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True, table_id=100
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)


def test_ipv6_dhcp_and_autoconf_ignore_gateway(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True, auto_gateway=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_extra_route)
    assert _poll(_has_ipv6_auto_nameserver)
    assert not _has_ipv6_auto_gateway()


def test_ipv6_dhcp_and_autoconf_ignore_dns(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True, auto_dns=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_ipv6_auto_extra_route)
    assert not _has_ipv6_auto_nameserver()


def test_ipv6_dhcp_and_autoconf_ignore_routes(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True, auto_routes=False
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_nameserver)
    assert not _has_ipv6_auto_gateway()
    assert not _has_ipv6_auto_extra_route()


def test_ipv4_dhcp_off_and_option_on(dhcpcli_up):
    """
    AUTO_ROUTES, AUTO_DNS and AUTO_GATEWAY should be silently ignored when
    DHCP is disabled.
    """
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    ipv4_state = _create_ipv4_state(
        enabled=True,
        dhcp=False,
        auto_dns=False,
        auto_gateway=False,
        auto_routes=False,
    )
    ipv4_state[InterfaceIPv4.ADDRESS] = [
        create_ipv4_address_state(IPV4_ADDRESS2, 24),
    ]

    dhcp_cli_desired_state[Interface.IPV4] = ipv4_state

    libnmstate.apply(desired_state)

    current_state = statelib.show_only((DHCP_CLI_NIC,))
    dhcp_cli_current_state = current_state[Interface.KEY][0]
    ipv4_current_state = dhcp_cli_current_state[Interface.IPV4]
    assert not ipv4_current_state[InterfaceIPv4.DHCP]
    assert InterfaceIPv4.AUTO_ROUTES not in ipv4_current_state
    assert InterfaceIPv4.AUTO_DNS not in ipv4_current_state
    assert InterfaceIPv4.AUTO_GATEWAY not in ipv4_current_state
    assert not _poll(_has_ipv4_dhcp_nameserver)
    assert not _has_ipv4_dhcp_gateway()
    assert not _has_ipv4_classless_route()


def test_ipv6_dhcp_off_and_option_on(dhcpcli_up):
    """
    AUTO_ROUTES, AUTO_DNS and AUTO_GATEWAY should be silently ignored when
    DHCP is disabled.
    """
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    ipv6_state = _create_ipv6_state(
        enabled=True,
        dhcp=False,
        autoconf=False,
        auto_dns=False,
        auto_gateway=False,
        auto_routes=False,
    )
    ipv6_state[InterfaceIPv6.ADDRESS] = [
        create_ipv6_address_state(IPV6_ADDRESS2, 64),
    ]
    dhcp_cli_desired_state[Interface.IPV6] = ipv6_state

    libnmstate.apply(desired_state)

    current_state = statelib.show_only((DHCP_CLI_NIC,))
    dhcp_cli_current_state = current_state[Interface.KEY][0]
    ipv6_current_state = dhcp_cli_current_state[Interface.IPV6]
    assert not ipv6_current_state[InterfaceIPv6.DHCP]
    assert InterfaceIPv6.AUTO_ROUTES not in ipv6_current_state
    assert InterfaceIPv6.AUTO_DNS not in ipv6_current_state
    assert InterfaceIPv6.AUTO_GATEWAY not in ipv6_current_state
    assert not _poll(_has_ipv6_auto_gateway)
    assert not _has_ipv6_auto_extra_route()
    assert not _has_ipv6_auto_nameserver()


def test_ipv4_dhcp_switch_on_to_off(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)

    # disable dhcp and make sure dns, route, gone.
    desired_state = statelib.show_only((DHCP_CLI_NIC,))
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(enabled=True)

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    assert not _poll(_has_ipv4_dhcp_nameserver)
    assert not _has_ipv4_dhcp_gateway()
    assert not _has_ipv4_classless_route()


def test_ipv6_dhcp_switch_on_to_off(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_ipv6_auto_extra_route)
    assert _poll(_has_ipv6_auto_nameserver)

    # disable dhcp and make sure dns, route, gone.
    desired_state = statelib.show_only((DHCP_CLI_NIC,))
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(enabled=True)

    libnmstate.apply(desired_state)

    assertlib.assert_state(desired_state)
    assert not _poll(_has_ipv6_auto_gateway)
    assert not _has_ipv6_auto_extra_route()
    assert not _has_ipv6_auto_nameserver()


@pytest.mark.tier1
def test_dhcp_on_bridge0(dhcpcli_up_with_dynamic_ip):
    """
    Test dynamic IPv4 & IPv6 addresses over a Linux bridge interface.

    Several checks have been ecooperated in the test due to the high time cost.
    The dynamic IP over the bridge includes the follwing checks:
        - The dynamic settings have been applied.
        - IPv4 and IPv6 addresses have been provided by the server.
        - IPv4 addresses are identical to the original ones which existed on
        the nic (dhcpcli interface).
        - IPv6 addresses are identical to the original ones which existed on
        the nic (dhcpcli interface).
    """
    origin_port_state = dhcpcli_up_with_dynamic_ip

    port_name = origin_port_state[Interface.KEY][0][Interface.NAME]

    bridge_state = create_bridge_subtree_state()
    bridge_state = add_port_to_bridge(bridge_state, port_name)

    bridge_iface_state = {
        Interface.IPV4: _create_ipv4_state(enabled=True, dhcp=True),
        Interface.IPV6: _create_ipv6_state(
            enabled=True, dhcp=True, autoconf=True
        ),
        Interface.MAC: get_mac_address(DHCP_CLI_NIC),
    }
    bridge_name = TEST_BRIDGE_NIC
    with linux_bridge(bridge_name, bridge_state, bridge_iface_state) as state:
        assertlib.assert_state_match(state)

        assert _poll(_has_dhcpv4_addr, nic=TEST_BRIDGE_NIC)
        assert _poll(_has_ipv4_dhcp_gateway, nic=TEST_BRIDGE_NIC)
        assert _poll(_has_dhcpv6_addr, nic=TEST_BRIDGE_NIC)
        assert _poll(_has_ipv6_auto_gateway, nic=TEST_BRIDGE_NIC)
        new_bridge_state = statelib.show_only((bridge_name,))

    new_ipv4_state = new_bridge_state[Interface.KEY][0][Interface.IPV4]
    new_ipv6_state = new_bridge_state[Interface.KEY][0][Interface.IPV6]
    assert new_ipv4_state[InterfaceIPv4.ADDRESS]
    assert len(new_ipv6_state[InterfaceIPv6.ADDRESS]) > 1

    origin_ipv4_state = origin_port_state[Interface.KEY][0][Interface.IPV4]
    origin_ipv6_state = origin_port_state[Interface.KEY][0][Interface.IPV6]
    _sort_ip_addresses(origin_ipv4_state[InterfaceIP.ADDRESS])
    _sort_ip_addresses(origin_ipv6_state[InterfaceIP.ADDRESS])
    _sort_ip_addresses(new_ipv4_state[InterfaceIP.ADDRESS])
    _sort_ip_addresses(new_ipv6_state[InterfaceIP.ADDRESS])
    assert origin_ipv4_state == new_ipv4_state
    assert origin_ipv6_state == new_ipv6_state


@pytest.mark.tier1
def test_port_ipaddr_learned_via_dhcp_added_as_static_to_linux_bridge(
    dhcpcli_up,
):
    dhcpcli_up[Interface.KEY][0][Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )

    libnmstate.apply(dhcpcli_up)

    assert _poll(_has_dhcpv4_addr)

    port_ifname = dhcpcli_up[Interface.KEY][0][Interface.NAME]
    port_state = statelib.show_only((port_ifname,))
    port_iface_state = port_state[Interface.KEY][0]
    dhcpcli_ip = port_iface_state[Interface.IPV4][InterfaceIPv4.ADDRESS]

    bridge_state = add_port_to_bridge(
        create_bridge_subtree_state(), port_ifname
    )

    ipv4_state = _create_ipv4_state(enabled=True, dhcp=False)
    ipv4_state[InterfaceIPv4.ADDRESS] = dhcpcli_ip
    with linux_bridge(
        TEST_BRIDGE_NIC,
        bridge_state,
        extra_iface_state={Interface.IPV4: ipv4_state},
        create=False,
    ) as state:
        state[Interface.KEY].append(
            {
                Interface.NAME: port_ifname,
                Interface.IPV4: _create_ipv4_state(enabled=False),
                Interface.IPV6: _create_ipv6_state(enabled=False),
            }
        )
        libnmstate.apply(state)

        assertlib.assert_state_match(state)


@pytest.mark.xfail(raises=NmstateNotImplementedError, strict=True)
def test_ipv6_autoconf_only(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, autoconf=True
    )

    libnmstate.apply(desired_state)


def _setup_dhcp_nics():
    cmdlib.exec_cmd(
        f"ip netns exec {DHCP_SRV_NS} "
        f"ip addr add {DHCP_SRV_IP4}/24 dev {DHCP_SRV_NIC}".split(),
        check=True,
    )
    # This stop dhcp server NIC get another IPv6 address from dnsmasq.
    cmdlib.exec_cmd(
        f"ip netns exec {DHCP_SRV_NS} "
        f"sysctl -w net.ipv6.conf.{DHCP_SRV_NIC}.accept_ra=0".split(),
        check=True,
    )

    cmdlib.exec_cmd(
        f"ip netns exec {DHCP_SRV_NS} "
        f"ip addr add {DHCP_SRV_IP6}/64 dev {DHCP_SRV_NIC}".split(),
        check=True,
    )

    cmdlib.exec_cmd(
        f"ip netns exec {DHCP_SRV_NS} "
        f"ip addr add {DHCP_SRV_IP6_2}/64 dev {DHCP_SRV_NIC}".split(),
        check=True,
    )


def _clean_up():
    dnsmasq_pid = cmdlib.exec_cmd(["pidof", "dnsmasq"])[1]
    cmdlib.exec_cmd(["kill", dnsmasq_pid.strip()])
    remove_veth_pair(DHCP_CLI_NIC, DHCP_SRV_NS)
    try:
        os.unlink(DNSMASQ_CONF_PATH)
    except (FileNotFoundError, OSError):
        pass


def _get_nameservers():
    """
    Return a list of name server string configured in RESOLV_CONF_PATH.
    """
    running_ns = (
        libnmstate.show()
        .get(Constants.DNS, {})
        .get(DNS.RUNNING, {})
        .get(DNS.SERVER, [])
    )
    logging.debug("Current running DNS: {}".format(running_ns))
    return running_ns


def _get_running_routes():
    """
    return a list of running routes
    """
    running_routes = (
        libnmstate.show().get(Constants.ROUTES, {}).get(RT.RUNNING, [])
    )
    logging.debug("Current running routes: {}".format(running_routes))
    return running_routes


def _poll(func, *args, **kwargs):
    return retry_till_true_or_timeout(DEFAULT_TIMEOUT, func, *args, **kwargs)


def _has_ipv6_auto_gateway(nic=DHCP_CLI_NIC):
    routes = _get_running_routes()
    for route in routes:
        if (
            route[RT.DESTINATION] == IPV6_DEFAULT_GATEWAY
            and route[RT.NEXT_HOP_INTERFACE] == nic
        ):
            return True
    return False


def _has_ipv6_auto_extra_route():
    routes = _get_running_routes()
    for route in routes:
        if (
            route[RT.DESTINATION] == IPV6_CLASSLESS_ROUTE_DST_NET1
            and route[RT.NEXT_HOP_INTERFACE] == DHCP_CLI_NIC
        ):
            return True
    return False


def _has_ipv6_auto_nameserver():
    return DHCP_SRV_IP6 in _get_nameservers()


def _has_ipv4_dhcp_nameserver():
    return DHCP_SRV_IP4 in _get_nameservers()


def _has_ipv4_dhcp_gateway(nic=DHCP_CLI_NIC):
    routes = _get_running_routes()
    for route in routes:
        if (
            route[RT.DESTINATION] == IPV4_DEFAULT_GATEWAY
            and route[RT.NEXT_HOP_INTERFACE] == nic
        ):
            return True
    return False


def _has_ipv4_classless_route():
    routes = _get_running_routes()
    for route in routes:
        if (
            route[RT.DESTINATION] == IPV4_CLASSLESS_ROUTE_DST_NET1
            and route[RT.NEXT_HOP_ADDRESS] == IPV4_CLASSLESS_ROUTE_NEXT_HOP1
            and route[RT.NEXT_HOP_INTERFACE] == DHCP_CLI_NIC
        ):
            return True
    return False


def _has_dhcpv6_addr(nic=DHCP_CLI_NIC):
    current_state = statelib.show_only((nic,))[Interface.KEY][0]
    has_dhcp_ip_addr = False
    addrs = current_state[Interface.IPV6].get(InterfaceIPv6.ADDRESS, [])
    logging.debug("Current IPv6 address of {}: {}".format(nic, addrs))
    for addr in addrs:
        if (
            addr[InterfaceIPv6.ADDRESS_PREFIX_LENGTH] == 128
            and DHCP_SRV_IP6_PREFIX in addr[InterfaceIPv6.ADDRESS_IP]
        ):
            has_dhcp_ip_addr = True
            break
    return has_dhcp_ip_addr


def _has_dhcpv4_addr(nic=DHCP_CLI_NIC):
    current_state = statelib.show_only((nic,))[Interface.KEY][0]
    has_dhcp_ip_addr = False
    addrs = current_state[Interface.IPV4].get(InterfaceIPv4.ADDRESS, [])
    logging.debug("Current IPv4 address of {}: {}".format(nic, addrs))
    for addr in addrs:
        if (
            addr[InterfaceIPv6.ADDRESS_PREFIX_LENGTH] == 24
            and DHCP_SRV_IP4_PREFIX in addr[InterfaceIPv4.ADDRESS_IP]
        ):
            has_dhcp_ip_addr = True
            break
    return has_dhcp_ip_addr


def _create_ipv4_state(
    enabled,
    dhcp=False,
    auto_dns=True,
    auto_gateway=True,
    auto_routes=True,
    table_id=0,
):
    state = {
        InterfaceIPv4.ENABLED: enabled,
        InterfaceIPv4.DHCP: dhcp,
        InterfaceIPv4.AUTO_DNS: auto_dns,
        InterfaceIPv4.AUTO_GATEWAY: auto_gateway,
        InterfaceIPv4.AUTO_ROUTES: auto_routes,
    }
    if dhcp:
        state[InterfaceIPv4.AUTO_ROUTE_TABLE_ID] = table_id

    return state


def _create_ipv6_state(
    enabled,
    dhcp=False,
    autoconf=False,
    auto_dns=True,
    auto_gateway=True,
    auto_routes=True,
    table_id=0,
):
    state = {
        InterfaceIPv6.ENABLED: enabled,
        InterfaceIPv6.DHCP: dhcp,
        InterfaceIPv6.AUTOCONF: autoconf,
        InterfaceIPv6.AUTO_DNS: auto_dns,
        InterfaceIPv6.AUTO_GATEWAY: auto_gateway,
        InterfaceIPv6.AUTO_ROUTES: auto_routes,
    }

    if dhcp or autoconf:
        state[InterfaceIPv6.AUTO_ROUTE_TABLE_ID] = table_id

    return state


def create_ipv4_address_state(address, prefix_length):
    return {
        InterfaceIPv4.ADDRESS_IP: address,
        InterfaceIPv4.ADDRESS_PREFIX_LENGTH: prefix_length,
    }


def create_ipv6_address_state(address, prefix_length):
    return {
        InterfaceIPv6.ADDRESS_IP: address,
        InterfaceIPv6.ADDRESS_PREFIX_LENGTH: prefix_length,
    }


@pytest.fixture(scope="function")
def dummy00():
    ifstate = {
        Interface.NAME: "dummy00",
        Interface.TYPE: InterfaceType.DUMMY,
        Interface.STATE: InterfaceState.UP,
    }
    libnmstate.apply({Interface.KEY: [ifstate]})
    yield ifstate
    ifstate[Interface.STATE] = InterfaceState.ABSENT
    libnmstate.apply({Interface.KEY: [ifstate]}, verify_change=False)


@parametrize_ip_ver
def test_activate_dummy_without_dhcp_service(ip_ver, dummy00):
    ifstate = dummy00
    if Interface.IPV4 in ip_ver:
        ifstate[Interface.IPV4] = _create_ipv4_state(enabled=True, dhcp=True)
    if Interface.IPV6 in ip_ver:
        ifstate[Interface.IPV6] = _create_ipv6_state(
            enabled=True, dhcp=True, autoconf=True
        )
    libnmstate.apply({Interface.KEY: [ifstate]})


@pytest.mark.tier1
def test_dummy_disable_ip_stack_with_on_going_dhcp(dummy00):
    ifstate = dummy00
    ifstate[Interface.IPV4] = _create_ipv4_state(enabled=True, dhcp=True)
    ifstate[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )
    libnmstate.apply({Interface.KEY: [ifstate]})
    ifstate[Interface.IPV4] = _create_ipv4_state(enabled=False)
    ifstate[Interface.IPV6] = _create_ipv6_state(enabled=False)
    libnmstate.apply({Interface.KEY: [ifstate]})


def test_dhcp4_with_static_ipv6(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )
    dhcp_cli_desired_state[Interface.IPV6] = {
        InterfaceIPv6.ENABLED: True,
        InterfaceIPv6.ADDRESS: [
            {
                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
            }
        ],
    }

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)


def test_dhcp6_and_autoconf_with_static_ipv4(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )
    dhcp_cli_desired_state[Interface.IPV4] = {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.ADDRESS: [
            {
                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS2,
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            }
        ],
    }

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)

    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_dhcpv6_addr)


@pytest.fixture(scope="function")
def dhcpcli_up_with_static_ip(dhcpcli_up):
    desired_state = dhcpcli_up
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.ADDRESS: [
            {
                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS2,
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            }
        ],
    }
    dhcp_cli_desired_state[Interface.IPV6] = {
        InterfaceIPv6.ENABLED: True,
        InterfaceIPv6.ADDRESS: [
            {
                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
            }
        ],
    }

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    yield desired_state


@pytest.mark.tier1
def test_change_static_to_dhcp4_with_disabled_ipv6(dhcpcli_up_with_static_ip):
    desired_state = dhcpcli_up_with_static_ip
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )
    dhcp_cli_desired_state[Interface.IPV6] = {InterfaceIPv6.ENABLED: False}

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)


@pytest.mark.tier1
def test_change_static_to_dhcp6_autoconf_with_disabled_ipv4(
    dhcpcli_up_with_static_ip,
):
    desired_state = dhcpcli_up_with_static_ip
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]

    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )
    dhcp_cli_desired_state[Interface.IPV4] = {InterfaceIPv4.ENABLED: False}

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)
    assert _poll(_has_dhcpv6_addr)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_ipv6_auto_nameserver)
    assert _poll(_has_ipv6_auto_extra_route)


@pytest.mark.tier1
@pytest.mark.slow
@parametrize_ip_ver
def test_dummy_existance_after_dhcp_timeout(ip_ver, dummy00):
    ifstate = dummy00
    if Interface.IPV4 in ip_ver:
        ifstate[Interface.IPV4] = _create_ipv4_state(enabled=True, dhcp=True)
    if Interface.IPV6 in ip_ver:
        ifstate[Interface.IPV6] = _create_ipv6_state(
            enabled=True, dhcp=True, autoconf=False
        )
    libnmstate.apply({Interface.KEY: [ifstate]})
    time.sleep(NM_DHCP_TIMEOUT_DEFAULT + 1)
    # NetworkManager by default remove virtual interface after DHCP timeout
    assertlib.assert_state({Interface.KEY: [ifstate]})


@pytest.mark.tier1
@pytest.mark.slow
def test_dummy_existance_after_ipv6_autoconf_timeout(dummy00):
    ifstate = dummy00
    ifstate[Interface.IPV4] = _create_ipv4_state(enabled=False)
    ifstate[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )
    libnmstate.apply({Interface.KEY: [ifstate]})
    time.sleep(NM_IPV6_AUTOCONF_TIMEOUT_DEFAULT + 1)

    # NetworkManager by default remove virtual interface after autoconf timeout
    # According to RFC 4861, autoconf(IPv6-RA) will instruct client to do
    # DHCPv6 or not. With autoconf timeout, DHCPv6 will not start.
    assertlib.assert_state({Interface.KEY: [ifstate]})


@pytest.fixture(scope="function")
def dhcpcli_up_with_static_ip_and_route(dhcpcli_up_with_static_ip):
    desired_state = dhcpcli_up_with_static_ip
    desired_state[RT.KEY] = {
        RT.CONFIG: [
            {
                RT.DESTINATION: IPV4_DEFAULT_GATEWAY,
                RT.NEXT_HOP_ADDRESS: DHCP_SRV_IP4,
                RT.NEXT_HOP_INTERFACE: DHCP_CLI_NIC,
            },
            {
                RT.DESTINATION: IPV4_NETWORK1,
                RT.NEXT_HOP_ADDRESS: DHCP_SRV_IP4,
                RT.NEXT_HOP_INTERFACE: DHCP_CLI_NIC,
            },
            {
                RT.DESTINATION: IPV6_DEFAULT_GATEWAY,
                RT.NEXT_HOP_ADDRESS: IPV6_ADDRESS3,
                RT.NEXT_HOP_INTERFACE: DHCP_CLI_NIC,
            },
            {
                RT.DESTINATION: IPV6_NETWORK1,
                RT.NEXT_HOP_ADDRESS: IPV6_ADDRESS3,
                RT.NEXT_HOP_INTERFACE: DHCP_CLI_NIC,
            },
        ]
    }

    libnmstate.apply(desired_state)
    yield desired_state


@pytest.mark.tier1
def test_static_ip_with_routes_switch_back_to_dynamic(
    dhcpcli_up_with_static_ip_and_route,
):
    desired_state = dhcpcli_up_with_static_ip_and_route
    desired_state.pop(RT.KEY)
    dhcp_cli_desired_state = desired_state[Interface.KEY][0]
    dhcp_cli_desired_state[Interface.STATE] = InterfaceState.UP
    dhcp_cli_desired_state[Interface.IPV4] = _create_ipv4_state(
        enabled=True, dhcp=True
    )
    dhcp_cli_desired_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=True
    )

    libnmstate.apply(desired_state)
    assertlib.assert_state(desired_state)

    assert _poll(_has_ipv4_dhcp_nameserver)
    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv4_classless_route)
    assert _poll(_has_dhcpv4_addr)
    assert _poll(_has_dhcpv6_addr)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_ipv6_auto_extra_route)
    assert _poll(_has_ipv6_auto_nameserver)

    current_config_routes = [
        route
        for route in libnmstate.show()[RT.KEY][RT.CONFIG]
        if route[RT.NEXT_HOP_INTERFACE] == DHCP_CLI_NIC
    ]
    assert not current_config_routes


@pytest.fixture(scope="function")
def eth1_with_dhcp6_no_dhcp_server():
    # Cannot depend on eth1_up fixture as the reproducer requires the
    # veth profile been created with DHCPv6 enabled.
    iface_state = {
        Interface.NAME: ETH1,
        Interface.TYPE: InterfaceType.ETHERNET,
        Interface.STATE: InterfaceState.UP,
    }
    iface_state[Interface.IPV4] = _create_ipv4_state(enabled=False)
    iface_state[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=False
    )
    libnmstate.apply({Interface.KEY: [iface_state]})
    try:
        yield iface_state
    finally:
        libnmstate.apply(
            {
                Interface.KEY: [
                    {
                        Interface.NAME: ETH1,
                        Interface.STATE: InterfaceState.ABSENT,
                    }
                ]
            },
            verify_change=False,
        )


def test_switch_from_dynamic_ip_without_dhcp_srv_to_static_ipv6(
    eth1_with_dhcp6_no_dhcp_server,
):
    iface_state = eth1_with_dhcp6_no_dhcp_server
    iface_state[Interface.IPV4] = {InterfaceIPv4.ENABLED: False}
    iface_state[Interface.IPV6] = {
        InterfaceIPv6.ENABLED: True,
        InterfaceIPv6.DHCP: False,
        InterfaceIPv6.AUTOCONF: False,
        InterfaceIPv6.ADDRESS: [
            {
                InterfaceIPv6.ADDRESS_IP: IPV6_ADDRESS2,
                InterfaceIPv6.ADDRESS_PREFIX_LENGTH: 64,
            }
        ],
    }
    libnmstate.apply({Interface.KEY: [iface_state]})
    assertlib.assert_state_match({Interface.KEY: [iface_state]})


@pytest.fixture
def dhcpcli_up_with_dns_cleanup(dhcpcli_up):
    yield dhcpcli_up
    libnmstate.apply({DNS.KEY: {DNS.CONFIG: {}}})


def test_dynamic_ip_with_static_dns(dhcpcli_up_with_dns_cleanup, clean_state):
    iface_state = {
        Interface.NAME: DHCP_CLI_NIC,
        Interface.STATE: InterfaceState.UP,
        Interface.IPV4: _create_ipv4_state(
            enabled=True, dhcp=True, auto_dns=False
        ),
        Interface.IPV6: _create_ipv6_state(
            enabled=True, dhcp=True, autoconf=True, auto_dns=False
        ),
    }
    dns_config = {
        DNS.CONFIG: {
            DNS.SERVER: [IPV6_DNS_NAMESERVER, IPV4_DNS_NAMESERVER],
            DNS.SEARCH: EXAMPLE_SEARCHES,
        }
    }
    desired_state = {Interface.KEY: [iface_state], DNS.KEY: dns_config}

    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)

    assert _poll(_has_ipv4_dhcp_gateway)
    assert _poll(_has_ipv6_auto_gateway)
    assert _poll(_has_dhcpv4_addr)
    assert _poll(_has_dhcpv6_addr)
    assert not _has_ipv4_dhcp_nameserver()
    assert not _has_ipv6_auto_nameserver()
    new_state = libnmstate.show()
    assert dns_config[DNS.CONFIG] == new_state[DNS.KEY][DNS.CONFIG]
    assert dns_config[DNS.CONFIG] == new_state[DNS.KEY][DNS.RUNNING]


@pytest.fixture(scope="function")
def clean_state():
    current_state = libnmstate.show()
    desired_state = deepcopy(current_state)
    for iface_state in desired_state[Interface.KEY]:
        if iface_state[Interface.IPV4][InterfaceIPv4.ENABLED]:
            iface_state[Interface.IPV4][InterfaceIPv4.AUTO_DNS] = False
            iface_state[Interface.IPV4][InterfaceIPv4.AUTO_ROUTES] = False
        if iface_state[Interface.IPV6][InterfaceIPv6.ENABLED]:
            iface_state[Interface.IPV6][InterfaceIPv6.AUTO_DNS] = False
            iface_state[Interface.IPV6][InterfaceIPv6.AUTO_ROUTES] = False

    libnmstate.apply(desired_state)
    try:
        yield
    finally:
        libnmstate.apply(current_state)


def _sort_ip_addresses(addresses):
    addresses.sort(key=itemgetter(InterfaceIP.ADDRESS_IP))


def test_enable_dhcp_with_no_server(dummy00):
    iface_info = dummy00
    iface_info[Interface.IPV4] = _create_ipv4_state(enabled=True, dhcp=True)
    iface_info[Interface.IPV6] = _create_ipv6_state(
        enabled=True, dhcp=True, autoconf=False
    )
    desired_state = {Interface.KEY: [iface_info]}
    libnmstate.apply(desired_state)
    assertlib.assert_state_match(desired_state)


def test_show_running_config_does_not_include_auto_config(
    dhcpcli_up_with_dynamic_ip,
):
    running_config = libnmstate.show_running_config()
    dhcpcli_iface_config = None
    for iface_config in running_config[Interface.KEY]:
        if iface_config[Interface.NAME] == DHCP_CLI_NIC:
            dhcpcli_iface_config = iface_config
            break
    nmstatectl_output = cmdlib.exec_cmd(
        ["nmstatectl", "show", "-r", DHCP_CLI_NIC, "--json"]
    )[1]
    nmstatectl_iface_state = json.loads(nmstatectl_output)[Interface.KEY][0]

    for iface_config in (dhcpcli_iface_config, nmstatectl_iface_state):
        assert iface_config[Interface.IPV4][InterfaceIPv4.DHCP]
        assert iface_config[Interface.IPV6][InterfaceIPv6.DHCP]
        assert iface_config[Interface.IPV6][InterfaceIPv6.AUTOCONF]
        assert not iface_config[Interface.IPV4][InterfaceIPv4.ADDRESS]
        ipv6_addresses = iface_config[Interface.IPV6][InterfaceIPv6.ADDRESS]
        assert len(ipv6_addresses) == 1
        assert is_ipv6_link_local_addr(
            ipv6_addresses[0][InterfaceIPv6.ADDRESS_IP],
            ipv6_addresses[0][InterfaceIPv6.ADDRESS_PREFIX_LENGTH],
        )
    assert not running_config[DNS.KEY][DNS.CONFIG]
    assert not running_config[RT.KEY][RT.CONFIG]



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from unittest import mock

from libnmstate import nm
from libnmstate.nm.bridge_port_vlan import nmstate_port_vlan_to_nm
from libnmstate.schema import LinuxBridge as LB


ACCESS_TAG = 4000


@pytest.fixture
def NM_mock():
    with mock.patch.object(nm.bridge_port_vlan, "NM") as m:
        yield m


def _get_port_vlan_ranged_trunks(min_vlan, max_vlan):
    return {
        LB.Port.Vlan.TrunkTags.ID_RANGE: {
            LB.Port.Vlan.TrunkTags.MIN_RANGE: min_vlan,
            LB.Port.Vlan.TrunkTags.MAX_RANGE: max_vlan,
        }
    }


def _port_vlan_trunk_tags_to_tag_list(trunk_tags, access_tag, enable_vlan):
    tag_list = []
    for trunk_tag in trunk_tags:
        single_tag = trunk_tag.get(LB.Port.Vlan.TrunkTags.ID)
        range_tags = trunk_tag.get(LB.Port.Vlan.TrunkTags.ID_RANGE, {})
        if single_tag:
            tag_list.append(single_tag)
        else:
            tag_list += range(
                range_tags.get(LB.Port.Vlan.TrunkTags.MIN_RANGE),
                range_tags.get(LB.Port.Vlan.TrunkTags.MAX_RANGE) + 1,
            )
    if enable_vlan:
        tag_list.append(access_tag)
    return tag_list


@pytest.mark.parametrize(
    "trunk_tags",
    [
        [],
        [{LB.Port.Vlan.TrunkTags.ID: 100}],
        [{LB.Port.Vlan.TrunkTags.ID: 100}, {LB.Port.Vlan.TrunkTags.ID: 101}],
        [_get_port_vlan_ranged_trunks(100, 199)],
        [
            _get_port_vlan_ranged_trunks(100, 199),
            _get_port_vlan_ranged_trunks(500, 1000),
        ],
        [
            {LB.Port.Vlan.TrunkTags.ID: 100},
            _get_port_vlan_ranged_trunks(500, 1000),
        ],
    ],
    ids=[
        "no-trunk-tags",
        "single-trunk-tag",
        "two-single-trunk-tags",
        "tag-ranges",
        "multiple-tag-ranges",
        "mixed-single-tag-and-tag-ranges",
    ],
)
@pytest.mark.parametrize(
    "is_native_vlan", [False, True], ids=["not-native-vlan", "native-vlan"]
)
def test_generate_nm_vlan_filtering_config(trunk_tags, is_native_vlan):
    vlan_config_info = {
        LB.Port.Vlan.ENABLE_NATIVE: is_native_vlan,
        LB.Port.Vlan.MODE: LB.Port.Vlan.Mode.TRUNK,
        LB.Port.Vlan.TRUNK_TAGS: trunk_tags,
    }

    if is_native_vlan:
        vlan_config_info[LB.Port.Vlan.TAG] = ACCESS_TAG

    nm_port_bridge_vlans = nmstate_port_vlan_to_nm(vlan_config_info)
    expected_trunk_tag_length = len(trunk_tags) + (1 if is_native_vlan else 0)
    assert expected_trunk_tag_length == len(nm_port_bridge_vlans)
    assert is_native_vlan == any(
        port_bridge_vlan.is_pvid() and port_bridge_vlan.is_untagged()
        for port_bridge_vlan in nm_port_bridge_vlans
    )
    desired_trunk_tags = _port_vlan_trunk_tags_to_tag_list(
        trunk_tags, ACCESS_TAG, is_native_vlan
    )
    _assert_port_settings_vlan_conf(nm_port_bridge_vlans, desired_trunk_tags)


def _assert_port_settings_vlan_conf(nm_port_vlans, desired_vlans):
    expected_vlans = set(desired_vlans)
    for nm_port_vlan in nm_port_vlans:
        _, min_vlan, max_vlan = nm_port_vlan.get_vid_range()
        for i in range(min_vlan, max_vlan + 1):
            expected_vlans.remove(i)
    assert not expected_vlans, "vlans: {} were not configured".format(
        expected_vlans
    )



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from unittest import mock

import pytest

import libnmstate.nm.dns as nm_dns
import libnmstate.nm.ipv4 as nm_ipv4
import libnmstate.nm.ipv6 as nm_ipv6
from libnmstate.ifaces import BaseIface
from libnmstate.dns import DnsState
from libnmstate.nm.common import NM
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv4
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceState
from libnmstate.schema import Route


TEST_IPV4_GATEWAY_IFACE = "eth1"
TEST_IPV6_GATEWAY_IFACE = "eth2"
TEST_STATIC_ROUTE_IFACE = "eth3"

TEST_IFACE1 = "eth4"


@pytest.fixture()
def client_mock():
    yield mock.MagicMock()


def _get_test_dns_v4():
    return {
        DnsState.PRIORITY_METADATA: 40,
        DNS.SERVER: ["8.8.8.8", "1.1.1.1"],
        DNS.SEARCH: ["example.org", "example.com"],
    }


def _get_test_dns_v6():
    return {
        DnsState.PRIORITY_METADATA: 40,
        DNS.SERVER: ["2001:4860:4860::8888", "2606:4700:4700::1111"],
        DNS.SEARCH: ["example.net", "example.edu"],
    }


parametrize_ip_ver = pytest.mark.parametrize(
    "nm_ip", [(nm_ipv4), (nm_ipv6)], ids=["ipv4", "ipv6"]
)


parametrize_ip_ver_dns = pytest.mark.parametrize(
    "nm_ip, get_test_dns_func",
    [(nm_ipv4, _get_test_dns_v4), (nm_ipv6, _get_test_dns_v6)],
    ids=["ipv4", "ipv6"],
)


@parametrize_ip_ver
def test_add_dns_empty(nm_ip):
    dns_conf = {}
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: dns_conf},
        base_con_profile=None,
    )

    _assert_dns(setting_ip, dns_conf)


@parametrize_ip_ver_dns
def test_add_dns(nm_ip, get_test_dns_func):
    dns_conf = get_test_dns_func()
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: dns_conf},
        base_con_profile=None,
    )

    _assert_dns(setting_ip, dns_conf)


@parametrize_ip_ver_dns
def test_add_dns_duplicate_server(nm_ip, get_test_dns_func):
    dns_conf = get_test_dns_func()
    dns_conf[DNS.SERVER] = [dns_conf[DNS.SERVER][0], dns_conf[DNS.SERVER][0]]
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: dns_conf},
        base_con_profile=None,
    )

    dns_conf[DNS.SERVER] = [dns_conf[DNS.SERVER][0]]
    _assert_dns(setting_ip, dns_conf)


@parametrize_ip_ver_dns
def test_add_dns_duplicate_search(nm_ip, get_test_dns_func):
    dns_conf = get_test_dns_func()
    dns_conf[DNS.SEARCH] = [dns_conf[DNS.SEARCH][0], dns_conf[DNS.SEARCH][0]]
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: dns_conf},
        base_con_profile=None,
    )

    dns_conf[DNS.SEARCH] = [dns_conf[DNS.SEARCH][0]]
    _assert_dns(setting_ip, dns_conf)


@parametrize_ip_ver_dns
def test_clear_dns(client_mock, nm_ip, get_test_dns_func):
    dns_conf = get_test_dns_func()
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: dns_conf},
        base_con_profile=None,
    )
    nm_simple_conn = NM.SimpleConnection.new()
    nm_simple_conn.add_setting(setting_ip)

    new_setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.DNS_METADATA: {}},
        base_con_profile=nm_simple_conn,
    )

    _assert_dns(new_setting_ip, {})


def test_get_dns_nameserver_duplicated(client_mock):
    dns_entry1 = mock_nm_dns_entry("eth1", "192.0.2.1", "example.org")
    dns_entry2 = mock_nm_dns_entry("eth2", "192.0.2.1", "example.org")
    client_mock.get_dns_configuration.return_value = [dns_entry1, dns_entry2]

    dns_state = nm_dns.get_running(client_mock)
    assert dns_state[DNS.SERVER] == ["192.0.2.1"]


def test_get_dns_domain_duplicated(client_mock):
    dns_entry1 = mock_nm_dns_entry("eth1", "192.0.2.1", "example.org")
    dns_entry2 = mock_nm_dns_entry("eth2", "192.0.2.1", "example.org")
    client_mock.get_dns_configuration.return_value = [dns_entry1, dns_entry2]

    dns_state = nm_dns.get_running(client_mock)
    assert dns_state[DNS.SEARCH] == ["example.org"]


def _assert_dns(setting_ip, dns_conf):
    assert setting_ip.props.dns == dns_conf.get(DNS.SERVER, [])
    assert setting_ip.props.dns_search == dns_conf.get(DNS.SEARCH, [])
    if dns_conf:
        priority = (
            dns_conf.get(
                DnsState.PRIORITY_METADATA, nm_dns.DEFAULT_DNS_PRIORITY
            )
            + nm_dns.DNS_PRIORITY_STATIC_BASE
        )
        assert setting_ip.props.dns_priority == priority


def _get_test_ipv4_gateway():
    return {
        Route.DESTINATION: "0.0.0.0/0",
        Route.METRIC: 200,
        Route.NEXT_HOP_ADDRESS: "192.0.2.1",
        Route.NEXT_HOP_INTERFACE: TEST_IPV4_GATEWAY_IFACE,
        Route.TABLE_ID: 54,
    }


def _get_test_ipv6_gateway():
    return {
        Route.DESTINATION: "::/0",
        Route.METRIC: 201,
        Route.NEXT_HOP_ADDRESS: "2001:db8:2::f",
        Route.NEXT_HOP_INTERFACE: TEST_IPV6_GATEWAY_IFACE,
        Route.TABLE_ID: 54,
    }


def _get_test_static_routes():
    return [
        {
            Route.DESTINATION: "2001:db8:3::1",
            Route.METRIC: 201,
            Route.NEXT_HOP_ADDRESS: "2001:db8:2::f",
            Route.NEXT_HOP_INTERFACE: TEST_STATIC_ROUTE_IFACE,
            Route.TABLE_ID: 54,
        },
        {
            Route.DESTINATION: "198.51.100.0/24",
            Route.METRIC: 201,
            Route.NEXT_HOP_ADDRESS: "192.0.2.1",
            Route.NEXT_HOP_INTERFACE: TEST_STATIC_ROUTE_IFACE,
            Route.TABLE_ID: 54,
        },
    ]


def _get_test_desired_state_static_gateway(families):
    state = {
        Route.KEY: {Route.CONFIG: _get_test_static_routes()},
        Interface.KEY: [
            {
                Interface.NAME: TEST_IPV4_GATEWAY_IFACE,
                Interface.STATE: InterfaceState.UP,
            },
            {
                Interface.NAME: TEST_IPV6_GATEWAY_IFACE,
                Interface.STATE: InterfaceState.UP,
            },
            {
                Interface.NAME: TEST_STATIC_ROUTE_IFACE,
                Interface.STATE: InterfaceState.UP,
            },
        ],
    }
    if Interface.IPV4 in families:
        state[Route.KEY][Route.CONFIG].append(_get_test_ipv4_gateway())
    if Interface.IPV6 in families:
        state[Route.KEY][Route.CONFIG].append(_get_test_ipv6_gateway())
    return state


def _get_test_desired_state_dynamic_ip_but_no_auto_dns(families):
    iface_state = {
        Interface.NAME: TEST_IFACE1,
        Interface.STATE: InterfaceState.UP,
        Interface.IPV4: {InterfaceIPv4.ENABLED: False},
        Interface.IPV6: {InterfaceIPv6.ENABLED: False},
    }
    if Interface.IPV4 in families:
        iface_state[Interface.IPV4][InterfaceIPv4.ENABLED] = True
        iface_state[Interface.IPV4][InterfaceIPv4.DHCP] = True
        iface_state[Interface.IPV4][InterfaceIPv4.AUTO_DNS] = False

    if Interface.IPV6 in families:
        iface_state[Interface.IPV6][InterfaceIPv6.ENABLED] = True
        iface_state[Interface.IPV6][InterfaceIPv6.DHCP] = True
        iface_state[Interface.IPV6][InterfaceIPv6.AUTOCONF] = True
        iface_state[Interface.IPV6][InterfaceIPv6.AUTO_DNS] = False

    return {Interface.KEY: [iface_state]}


class mock_nm_dns_entry:
    def __init__(self, iface_name, nameserver, domain):
        self.iface_name = iface_name
        self.nameserver = nameserver
        self.domain = domain

    def get_interface(self):
        return self.iface_name

    def get_nameservers(self):
        return [self.nameserver]

    def get_domains(self):
        return [self.domain]



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import contextlib
import logging
import warnings

from libnmstate.error import NmstateValueError
from libnmstate.schema import Bond
from libnmstate.schema import BondMode
from libnmstate.schema import Interface

from ..state import state_match
from .base_iface import BaseIface


DEPRECATED_SLAVES = "slaves"


class BondIface(BaseIface):
    _MODE_CHANGE_METADATA = "_bond_mode_changed"

    def sort_port(self):
        if self.port:
            self.raw[Bond.CONFIG_SUBTREE][Bond.PORT].sort()

    def __init__(self, info, save_to_disk=True):
        super().__init__(info, save_to_disk)
        self._normalize_options_values()
        self._fix_bond_option_arp_monitor()
        self._replace_deprecated_terms()

    @property
    def port(self):
        return self.raw.get(Bond.CONFIG_SUBTREE, {}).get(Bond.PORT, [])

    @property
    def is_controller(self):
        return True

    @property
    def is_virtual(self):
        return True

    @property
    def bond_mode(self):
        return self.raw.get(Bond.CONFIG_SUBTREE, {}).get(Bond.MODE)

    @property
    def bond_options(self):
        return self.raw.get(Bond.CONFIG_SUBTREE, {}).get(
            Bond.OPTIONS_SUBTREE, {}
        )

    @property
    def is_bond_mode_changed(self):
        return self.raw.get(BondIface._MODE_CHANGE_METADATA) is True

    def _set_bond_mode_changed_metadata(self, value):
        self.raw[BondIface._MODE_CHANGE_METADATA] = value

    def _generate_bond_mode_change_metadata(self, ifaces):
        if self.is_up:
            cur_iface = ifaces.get_cur_iface(self.name, self.type)
            if cur_iface and self.bond_mode != cur_iface.bond_mode:
                self._set_bond_mode_changed_metadata(True)

    def gen_metadata(self, ifaces):
        super().gen_metadata(ifaces)
        if not self.is_absent:
            self._generate_bond_mode_change_metadata(ifaces)

    def pre_edit_validation_and_cleanup(self):
        super().pre_edit_validation_and_cleanup()
        if self.is_up:
            self._discard_bond_option_when_mode_change()
            self._validate_bond_mode()
            self._fix_mac_restriced_mode()
            self._validate_miimon_conflict_with_arp_interval()

    def _discard_bond_option_when_mode_change(self):
        if self.is_bond_mode_changed:
            logging.warning(
                "Discarding all current bond options as interface "
                f"{self.name} has bond mode changed"
            )
            self.raw[Bond.CONFIG_SUBTREE][
                Bond.OPTIONS_SUBTREE
            ] = self.original_dict.get(Bond.CONFIG_SUBTREE, {}).get(
                Bond.OPTIONS_SUBTREE, {}
            )
            self._normalize_options_values()

    def _validate_bond_mode(self):
        if self.bond_mode is None:
            raise NmstateValueError(
                f"Bond interface {self.name} does not have bond mode defined"
            )

    def _fix_mac_restriced_mode(self):
        if self.is_in_mac_restricted_mode:
            if self.original_dict.get(Interface.MAC):
                raise NmstateValueError(
                    "MAC address cannot be specified in bond interface along "
                    "with fail_over_mac active on active backup mode"
                )
            else:
                self.raw.pop(Interface.MAC, None)

    def _validate_miimon_conflict_with_arp_interval(self):
        bond_options = self.bond_options
        if bond_options.get("miimon") and bond_options.get("arp_interval"):
            raise NmstateValueError(
                "Bond option arp_interval is conflicting with miimon, "
                "please disable one of them by setting to 0"
            )

    def _replace_deprecated_terms(self):
        bond_cfg = self.raw.get(Bond.CONFIG_SUBTREE)
        if bond_cfg and bond_cfg.get(DEPRECATED_SLAVES):
            bond_cfg[Bond.PORT] = bond_cfg.pop(DEPRECATED_SLAVES)
            warnings.warn("Using 'slaves' is deprecated, use 'port' instead.")

    @staticmethod
    def is_mac_restricted_mode(mode, bond_options):
        return (
            mode == BondMode.ACTIVE_BACKUP
            and bond_options.get("fail_over_mac") == "active"
        )

    @property
    def is_in_mac_restricted_mode(self):
        """
        Return True when Bond option does not allow MAC address defined.
        In MAC restricted mode means:
            Bond mode is BondMode.ACTIVE_BACKUP
            Bond option "fail_over_mac" is active.
        """
        return BondIface.is_mac_restricted_mode(
            self.bond_mode, self.bond_options
        )

    def _normalize_options_values(self):
        if self.bond_options:
            normalized_options = {}
            for option_name, option_value in self.bond_options.items():
                with contextlib.suppress(ValueError):
                    option_value = int(option_value)
                option_value = _get_bond_named_option_value_by_id(
                    option_name, option_value
                )
                normalized_options[option_name] = option_value
            self.bond_options.update(normalized_options)

    def _fix_bond_option_arp_monitor(self):
        """
        Adding 'arp_ip_target=""' when ARP monitor is disabled by
        `arp_interval=0`
        """
        if self.bond_options:
            _include_arp_ip_target_explictly_when_disable(
                self.raw[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE]
            )

    def state_for_verify(self):
        state = super().state_for_verify()
        if state.get(Bond.CONFIG_SUBTREE, {}).get(Bond.OPTIONS_SUBTREE):
            _include_arp_ip_target_explictly_when_disable(
                state[Bond.CONFIG_SUBTREE][Bond.OPTIONS_SUBTREE]
            )
        return state

    def remove_port(self, port_name):
        self.raw[Bond.CONFIG_SUBTREE][Bond.PORT] = [
            s for s in self.port if s != port_name
        ]
        self.sort_port()

    def match_ignore_bond_options(self, current):
        self_state = self.state_for_verify()
        self_bond_options = self_state.get(Bond.CONFIG_SUBTREE, {}).pop(
            Bond.OPTIONS_SUBTREE, {}
        )
        current_state = current.state_for_verify()
        current_bond_options = current_state.get(Bond.CONFIG_SUBTREE, {}).pop(
            Bond.OPTIONS_SUBTREE, {}
        )
        for key, value in self_bond_options.items():
            current_value = current_bond_options.get(key)
            if current_value != value:
                if current_value is None:
                    logging.warning(
                        f"Desire iface {self.name} bond option {key}={value} "
                        f"is invalid in kernel for bond mode {self.bond_mode}"
                    )
                else:
                    logging.warning(
                        f"Desire iface {self.name} bond option {key}={value} "
                        f"does not match with kernel value: {current_value}"
                    )
        return state_match(self_state, current_state)


class _BondNamedOptions:
    AD_SELECT = "ad_select"
    ARP_ALL_TARGETS = "arp_all_targets"
    ARP_VALIDATE = "arp_validate"
    FAIL_OVER_MAC = "fail_over_mac"
    LACP_RATE = "lacp_rate"
    MODE = "mode"
    PRIMARY_RESELECT = "primary_reselect"
    XMIT_HASH_POLICY = "xmit_hash_policy"
    USE_CARRIER = "use_carrier"


_BOND_OPTIONS_NUMERIC_TO_NAMED_MAP = {
    _BondNamedOptions.AD_SELECT: ("stable", "bandwidth", "count"),
    _BondNamedOptions.ARP_ALL_TARGETS: ("any", "all"),
    _BondNamedOptions.ARP_VALIDATE: (
        "none",
        "active",
        "backup",
        "all",
        "filter",
        "filter_active",
        "filter_backup",
    ),
    _BondNamedOptions.FAIL_OVER_MAC: ("none", "active", "follow"),
    _BondNamedOptions.LACP_RATE: ("slow", "fast"),
    _BondNamedOptions.MODE: (
        "balance-rr",
        "active-backup",
        "balance-xor",
        "broadcast",
        "802.3ad",
        "balance-tlb",
        "balance-alb",
    ),
    _BondNamedOptions.PRIMARY_RESELECT: ("always", "better", "failure"),
    _BondNamedOptions.USE_CARRIER: (False, True),
    _BondNamedOptions.XMIT_HASH_POLICY: (
        "layer2",
        "layer3+4",
        "layer2+3",
        "encap2+3",
        "encap3+4",
    ),
}


def _get_bond_named_option_value_by_id(option_name, option_id_value):
    """
    Given an option name and its value, return a named option value
    if it exists.
    Return the same option value as inputted if:
    - The option name has no dual named and id values.
    - The option value is not numeric.
    - The option value has no corresponding named value (not in range).
    """
    option_value = _BOND_OPTIONS_NUMERIC_TO_NAMED_MAP.get(option_name)
    if option_value:
        with contextlib.suppress(ValueError, IndexError):
            return option_value[int(option_id_value)]
    return option_id_value


def _include_arp_ip_target_explictly_when_disable(bond_options):
    if (
        bond_options.get("arp_interval") == 0
        and "arp_ip_target" not in bond_options
    ):
        bond_options["arp_ip_target"] = ""



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import pytest

from unittest import mock

from libnmstate import nm
from libnmstate.schema import InterfaceIPv4

IPV4_ADDRESS1 = "192.0.2.251"


@pytest.fixture
def NM_mock():
    with mock.patch.object(nm.ipv4, "NM") as m:
        yield m


def test_create_setting_without_config(NM_mock):
    ipv4_setting = nm.ipv4.create_setting(config=None, base_con_profile=None)

    assert ipv4_setting == NM_mock.SettingIP4Config.new.return_value
    assert (
        ipv4_setting.props.method == NM_mock.SETTING_IP4_CONFIG_METHOD_DISABLED
    )


def test_create_setting_with_ipv4_disabled(NM_mock):
    ipv4_setting = nm.ipv4.create_setting(
        config={InterfaceIPv4.ENABLED: False}, base_con_profile=None
    )

    assert (
        ipv4_setting.props.method == NM_mock.SETTING_IP4_CONFIG_METHOD_DISABLED
    )


def test_create_setting_without_addresses(NM_mock):
    ipv4_setting = nm.ipv4.create_setting(
        config={InterfaceIPv4.ENABLED: True, InterfaceIPv4.ADDRESS: []},
        base_con_profile=None,
    )

    assert (
        ipv4_setting.props.method == NM_mock.SETTING_IP4_CONFIG_METHOD_DISABLED
    )


def test_create_setting_with_static_addresses(NM_mock):
    config = {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.ADDRESS: [
            {
                InterfaceIPv4.ADDRESS_IP: "10.10.10.1",
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            },
            {
                InterfaceIPv4.ADDRESS_IP: "10.10.20.1",
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            },
        ],
    }
    ipv4_setting = nm.ipv4.create_setting(config=config, base_con_profile=None)

    assert (
        ipv4_setting.props.method == NM_mock.SETTING_IP4_CONFIG_METHOD_MANUAL
    )
    NM_mock.IPAddress.new.assert_has_calls(
        [
            mock.call(
                nm.ipv4.socket.AF_INET,
                config[InterfaceIPv4.ADDRESS][0][InterfaceIPv4.ADDRESS_IP],
                config[InterfaceIPv4.ADDRESS][0][
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH
                ],
            ),
            mock.call(
                nm.ipv4.socket.AF_INET,
                config[InterfaceIPv4.ADDRESS][1][InterfaceIPv4.ADDRESS_IP],
                config[InterfaceIPv4.ADDRESS][1][
                    InterfaceIPv4.ADDRESS_PREFIX_LENGTH
                ],
            ),
        ]
    )
    NM_mock.SettingIP4Config.new.return_value.add_address.assert_has_calls(
        [
            mock.call(NM_mock.IPAddress.new.return_value),
            mock.call(NM_mock.IPAddress.new.return_value),
        ]
    )


def test_get_info_with_no_applied_config():
    info = nm.ipv4.get_ipv4_config(None)
    assert info == {}


def test_get_info_with_no_ip_profile():
    applied_config_mock = mock.MagicMock()
    applied_config_mock.get_setting_ip4_config.return_value = None

    info = nm.ipv4.get_ipv4_config(applied_config_mock)

    assert info == {InterfaceIPv4.ENABLED: False, InterfaceIPv4.DHCP: False}


def test_get_info_with_ip_profile(NM_mock):
    applied_config_mock = mock.MagicMock()
    ip_profile = mock.MagicMock()
    applied_config_mock.get_setting_ip4_config.return_value = ip_profile
    ip_profile.get_method.return_value = (
        NM_mock.SETTING_IP4_CONFIG_METHOD_MANUAL
    )
    ip_profile.props.never_default = False
    ip_profile.props.ignore_auto_dns = False
    ip_profile.props.ignore_auto_routes = False

    info = nm.ipv4.get_ipv4_config(applied_config_mock)

    assert info == {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.DHCP: False,
    }


def test_create_setting_with_base_con_profile(NM_mock):
    config = {
        InterfaceIPv4.ENABLED: True,
        InterfaceIPv4.ADDRESS: [
            {
                InterfaceIPv4.ADDRESS_IP: IPV4_ADDRESS1,
                InterfaceIPv4.ADDRESS_PREFIX_LENGTH: 24,
            }
        ],
    }
    base_con_profile_mock = mock.MagicMock()
    config_mock = base_con_profile_mock.get_setting_ip4_config.return_value
    config_dup_mock = config_mock.duplicate.return_value

    nm.ipv4.create_setting(
        config=config, base_con_profile=base_con_profile_mock
    )

    base_con_profile_mock.get_setting_ip4_config.assert_called_once_with()
    config_mock.duplicate.assert_called_once_with()
    config_dup_mock.clear_addresses.assert_called_once_with()



#
# Copyright (c) 2019-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy
from unittest import mock

import pytest

from libnmstate.nm import ipv4 as nm_ipv4
from libnmstate.nm import ipv6 as nm_ipv6
from libnmstate.nm.common import NM
from libnmstate.ifaces import BaseIface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import Route

IPV4_DEFAULT_GATEWAY_DESTINATION = "0.0.0.0/0"
IPV6_DEFAULT_GATEWAY_DESTINATION = "::/0"

IPV4_ROUTE1 = {
    Route.DESTINATION: "198.51.100.0/24",
    Route.METRIC: 103,
    Route.NEXT_HOP_ADDRESS: "192.0.2.1",
    Route.TABLE_ID: 50,
}

IPV4_ROUTE2 = {
    Route.DESTINATION: "203.0.113.0/24",
    Route.METRIC: 103,
    Route.NEXT_HOP_ADDRESS: "192.0.2.2",
    Route.TABLE_ID: 51,
}

IPV6_ROUTE1 = {
    Route.DESTINATION: "2001:db8:a::/64",
    Route.METRIC: 103,
    Route.NEXT_HOP_ADDRESS: "2001:db8:1::a",
    Route.TABLE_ID: 50,
}

IPV6_ROUTE2 = {
    Route.DESTINATION: "2001:db8:b::/64",
    Route.METRIC: 103,
    Route.NEXT_HOP_ADDRESS: "2001:db8:1::b",
    Route.TABLE_ID: 51,
}

parametrize_ip_ver_routes = pytest.mark.parametrize(
    "nm_ip, routes",
    [
        (nm_ipv4, [IPV4_ROUTE1, IPV4_ROUTE2]),
        (nm_ipv6, [IPV6_ROUTE1, IPV6_ROUTE2]),
    ],
    ids=["ipv4", "ipv6"],
)


@pytest.fixture
def client_mock():
    yield mock.MagicMock()


def _get_test_ipv4_gateways():
    return [
        {
            Route.DESTINATION: "0.0.0.0/0",
            Route.METRIC: 103,
            Route.NEXT_HOP_ADDRESS: "192.0.2.1",
            Route.TABLE_ID: 52,
        },
        {
            Route.DESTINATION: "0.0.0.0/0",
            Route.METRIC: 101,
            Route.NEXT_HOP_ADDRESS: "192.0.2.2",
            Route.TABLE_ID: 53,
        },
    ]


def _get_test_ipv6_gateways():
    return [
        {
            Route.DESTINATION: "::/0",
            Route.METRIC: 103,
            Route.NEXT_HOP_ADDRESS: "2001:db8:1::f",
            Route.TABLE_ID: 52,
        },
        {
            Route.DESTINATION: "::/0",
            Route.METRIC: 101,
            Route.NEXT_HOP_ADDRESS: "2001:db8:1::e",
            Route.TABLE_ID: 53,
        },
    ]


parametrize_ip_ver_routes_gw = pytest.mark.parametrize(
    "nm_ip, routes, gateways",
    [
        (nm_ipv4, [IPV4_ROUTE1, IPV4_ROUTE2], _get_test_ipv4_gateways()),
        (nm_ipv6, [IPV6_ROUTE1, IPV6_ROUTE2], _get_test_ipv6_gateways()),
    ],
    ids=["ipv4", "ipv6"],
)


@parametrize_ip_ver_routes
def test_add_multiple_route(nm_ip, routes):
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.ROUTES_METADATA: routes},
        base_con_profile=None,
    )
    assert [_nm_route_to_dict(r) for r in setting_ip.props.routes] == routes


@parametrize_ip_ver_routes
def test_add_duplicate_routes(nm_ip, routes):
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: [routes[0], routes[0]],
        },
        base_con_profile=None,
    )
    assert [_nm_route_to_dict(r) for r in setting_ip.props.routes] == [
        routes[0]
    ]


@parametrize_ip_ver_routes
def test_clear_route(nm_ip, routes, client_mock):
    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.ROUTES_METADATA: routes},
        base_con_profile=None,
    )
    nm_simple_conn = NM.SimpleConnection.new()
    nm_simple_conn.add_setting(setting_ip)

    new_setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.ROUTES_METADATA: []},
        base_con_profile=nm_simple_conn,
    )
    assert not [_nm_route_to_dict(r) for r in new_setting_ip.props.routes]


@parametrize_ip_ver_routes
def test_add_route_without_metric(nm_ip, routes):
    route_with_default_metric = copy.deepcopy(routes[0])
    route_with_default_metric[Route.METRIC] = Route.USE_DEFAULT_METRIC
    route_without_metric = copy.deepcopy(routes[0])
    del route_without_metric[Route.METRIC]
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: [route_without_metric],
        },
        base_con_profile=None,
    )
    assert [_nm_route_to_dict(r) for r in setting_ip.props.routes] == [
        route_with_default_metric
    ]


@parametrize_ip_ver_routes
def test_add_route_without_table_id(nm_ip, routes):
    route_with_default_table_id = copy.deepcopy(routes[0])
    route_with_default_table_id[Route.TABLE_ID] = Route.USE_DEFAULT_ROUTE_TABLE
    route_without_table_id = copy.deepcopy(routes[0])
    del route_without_table_id[Route.TABLE_ID]
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: [route_without_table_id],
        },
        base_con_profile=None,
    )
    assert [_nm_route_to_dict(r) for r in setting_ip.props.routes] == [
        route_with_default_table_id
    ]


@parametrize_ip_ver_routes_gw
def test_change_gateway(nm_ip, routes, gateways):
    desired_routes = routes + gateways[:1]
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: desired_routes,
        },
        base_con_profile=None,
    )
    assert [
        _nm_route_to_dict(r) for r in setting_ip.props.routes
    ] == routes + gateways[:1]
    assert _get_gateways(setting_ip)[0] == gateways[0]


@parametrize_ip_ver_routes_gw
def test_add_two_gateway(nm_ip, routes, gateways):
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: routes + gateways,
        },
        base_con_profile=None,
    )
    assert _get_gateways(setting_ip) == gateways


@parametrize_ip_ver_routes_gw
def test_add_duplicate_gateways(nm_ip, routes, gateways):
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: routes + [gateways[0], gateways[0]],
        },
        base_con_profile=None,
    )
    assert _get_gateways(setting_ip)[0] == gateways[0]


@parametrize_ip_ver_routes_gw
def test_change_gateway_without_metric(nm_ip, routes, gateways):
    del gateways[0][Route.METRIC]
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: routes + [gateways[0]],
        },
        base_con_profile=None,
    )
    current_routes = [_nm_route_to_dict(r) for r in setting_ip.props.routes]
    gateways[0][Route.METRIC] = Route.USE_DEFAULT_METRIC
    assert current_routes == routes + [gateways[0]]
    assert _get_gateways(setting_ip) == [gateways[0]]


@parametrize_ip_ver_routes_gw
def test_change_gateway_without_table_id(nm_ip, routes, gateways):
    del gateways[0][Route.TABLE_ID]
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: routes + [gateways[0]],
        },
        base_con_profile=None,
    )
    gateways[0][Route.TABLE_ID] = Route.USE_DEFAULT_ROUTE_TABLE

    assert [
        _nm_route_to_dict(r) for r in setting_ip.props.routes
    ] == routes + [gateways[0]]
    assert _get_gateways(setting_ip)[0] == gateways[0]


@parametrize_ip_ver_routes_gw
def test_clear_gateway(nm_ip, routes, gateways, client_mock):
    setting_ip = nm_ip.create_setting(
        {
            InterfaceIP.ENABLED: True,
            BaseIface.ROUTES_METADATA: routes + gateways[:1],
        },
        base_con_profile=None,
    )
    nm_simple_conn = NM.SimpleConnection.new()
    nm_simple_conn.add_setting(setting_ip)

    setting_ip = nm_ip.create_setting(
        {InterfaceIP.ENABLED: True, BaseIface.ROUTES_METADATA: routes},
        base_con_profile=nm_simple_conn,
    )
    assert [_nm_route_to_dict(r) for r in setting_ip.props.routes] == routes
    assert not setting_ip.get_gateway()
    assert setting_ip.get_route_table() == Route.USE_DEFAULT_ROUTE_TABLE
    assert setting_ip.props.route_metric == Route.USE_DEFAULT_METRIC


def _nm_route_to_dict(nm_route):
    dst = "{ip}/{prefix}".format(
        ip=nm_route.get_dest(), prefix=nm_route.get_prefix()
    )
    next_hop = nm_route.get_next_hop() or ""
    metric = int(nm_route.get_metric())
    table_id_variant = nm_route.get_attribute(NM.IP_ROUTE_ATTRIBUTE_TABLE)

    return {
        Route.TABLE_ID: int(table_id_variant.get_uint32()),
        Route.DESTINATION: dst,
        Route.NEXT_HOP_ADDRESS: next_hop,
        Route.METRIC: metric,
    }


def _get_gateways(setting_ip):
    gateways = []

    for nm_route in setting_ip.props.routes:
        route = _nm_route_to_dict(nm_route)
        if route[Route.DESTINATION] in (
            IPV6_DEFAULT_GATEWAY_DESTINATION,
            IPV4_DEFAULT_GATEWAY_DESTINATION,
        ):
            gateways.append(route)
    return gateways



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager
import os
import time
import yaml

import pytest

import libnmstate
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import LLDP

from .testlib import assertlib
from .testlib import cmdlib
from .testlib import ifacelib
from .testlib import statelib
from .testlib.veth import create_veth_pair
from .testlib.veth import remove_veth_pair


LLDPTEST = "lldptest"
LLDPTEST_PEER = "lldptest.peer"

LLDP_TEST_NS = "nmstate_lldp_test"

LLDP_SYSTEM_DESC = (
    "Summit300-48 - Version 7.4e.1 (Build 5) by Release_Master "
    "05/27/05 04:53:11"
)

EXPECTED_LLDP_NEIGHBOR = """
- system-name: Summit300-48
  type: 5
- system-description: Summit300-48 - Version 7.4e.1 (Build 5) by Release_Master
    05/27/05 04:53:11
  type: 6
- system-capabilities:
  - MAC Bridge component
  - Router
  type: 7
- _description: MAC address
  chassis-id: 00:01:30:F9:AD:A0
  chassis-id-type: 4
  type: 1
- _description: Interface name
  port-id: 1/1
  port-id-type: 5
  type: 2
- ieee-802-1-vlans:
  - name: v2-0488-03-0505
    vid: 488
  oui: 00:80:c2
  subtype: 3
  type: 127
- ieee-802-3-mac-phy-conf:
    autoneg: true
    operational-mau-type: 16
    pmd-autoneg-cap: 27648
  oui: 00:12:0f
  subtype: 1
  type: 127
- ieee-802-1-ppvids:
  - 0
  oui: 00:80:c2
  subtype: 2
  type: 127
- management-addresses:
  - address: 00:01:30:F9:AD:A0
    address-subtype: MAC
    interface-number: 1001
    interface-number-subtype: 2
  type: 8
- ieee-802-3-max-frame-size: 1522
  oui: 00:12:0f
  subtype: 4
  type: 127
"""

LLDP_CAPS = ["MAC Bridge component", "Router"]

CHASSIS_ID = "chassis-id"
CHASSIS_ID_TYPE = "chassis-id-type"
MANAGEMENT_ADDRESSES = "management-addresses"
ADDRESS = "address"
ADDRESS_SUBTYPE = "address-subtype"
INTERFACE_NUMBER = "interface-number"
INTERFACE_NUMBER_SUBTYPE = "interface-number-subtype"
AUTONEG = "autoneg"
OPERATIONAL_MAU_TYPE = "operational-mau-type"
PMD_AUTONEG_CAP = "pmd-autoneg-cap"
PPVIDS_SUBTREE = "ieee-802-1-ppvids"
PPVID = "ppvid"
VLANS_SUBTREE = "ieee-802-1-vlans"
NAME = "name"
PORT_ID = "port-id"
PORT_ID_TYPE = "port-id-type"
SYSTEM_NAME = "system-name"
SYSTEM_DESCRIPTION = "system-description"
SYSTEM_CAPABILITIES = "system-capabilities"
MFS_KEY = "ieee-802-3-max-frame-size"
MAC_PHY_SUBTREE = "ieee-802-3-mac-phy-conf"

LLDP_TEST_SYSTEM_NAME = "Summit300-48"


@pytest.fixture(scope="module")
def lldpiface_env():
    try:
        create_veth_pair(LLDPTEST, LLDPTEST_PEER, LLDP_TEST_NS)
        yield
    finally:
        remove_veth_pair(LLDPTEST, LLDP_TEST_NS)


@pytest.fixture
def lldptest_up(lldpiface_env):
    with ifacelib.iface_up(LLDPTEST) as ifstate:
        yield ifstate


def test_enable_lldp(lldptest_up):
    with lldp_enabled(lldptest_up) as dstate:
        assertlib.assert_state_match(dstate)


def test_lldp_yaml(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]
        assert test_neighbor == yaml.safe_load(EXPECTED_LLDP_NEIGHBOR)


def test_lldp_system(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        seen = set()
        for tlv in test_neighbor:
            seen.add(tlv[LLDP.Neighbors.TLV_TYPE])
            if tlv[LLDP.Neighbors.TLV_TYPE] == 5:
                assert tlv[SYSTEM_NAME] == "Summit300-48"
            elif tlv[LLDP.Neighbors.TLV_TYPE] == 6:
                assert tlv[SYSTEM_DESCRIPTION] == LLDP_SYSTEM_DESC
            elif tlv[LLDP.Neighbors.TLV_TYPE] == 7:
                assert tlv[SYSTEM_CAPABILITIES] == LLDP_CAPS
        assert set([5, 6, 7]).issubset(seen)


def test_lldp_chassis(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 1, test_neighbor
            )
        )
        assert len(tlvs) == 1
        assert tlvs[0][CHASSIS_ID] == "00:01:30:F9:AD:A0"
        assert tlvs[0][CHASSIS_ID_TYPE] == 4


def test_lldp_management_addresses(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 8, test_neighbor
            )
        )
        assert len(tlvs) == 1
        test_mngt = tlvs[0][MANAGEMENT_ADDRESSES][0]
        assert test_mngt[ADDRESS] == "00:01:30:F9:AD:A0"
        assert test_mngt[ADDRESS_SUBTYPE] == "MAC"
        assert test_mngt[INTERFACE_NUMBER] == 1001
        assert test_mngt[INTERFACE_NUMBER_SUBTYPE] == 2


def test_lldp_macphy(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 127
                and tlv[LLDP.Neighbors.ORGANIZATION_CODE] == "00:12:0f"
                and tlv[LLDP.Neighbors.TLV_SUBTYPE] == 1,
                test_neighbor,
            )
        )
        assert len(tlvs) == 1
        assert tlvs[0][MAC_PHY_SUBTREE][AUTONEG] is True
        assert tlvs[0][MAC_PHY_SUBTREE][OPERATIONAL_MAU_TYPE] == 16
        assert tlvs[0][MAC_PHY_SUBTREE][PMD_AUTONEG_CAP] == 27648


def test_lldp_port(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 2, test_neighbor
            )
        )
        assert len(tlvs) == 1
        assert tlvs[0][PORT_ID] == "1/1"
        assert tlvs[0][PORT_ID_TYPE] == 5
        assert tlvs[0][LLDP.Neighbors.DESCRIPTION] == "Interface name"


def test_lldp_port_vlan(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 127
                and tlv[LLDP.Neighbors.ORGANIZATION_CODE] == "00:80:c2"
                and tlv[LLDP.Neighbors.TLV_SUBTYPE] == 2,
                test_neighbor,
            )
        )
        assert len(tlvs) == 1
        assert tlvs[0][PPVIDS_SUBTREE][0] == 0


def test_lldp_vlan(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 127
                and tlv[LLDP.Neighbors.ORGANIZATION_CODE] == "00:80:c2"
                and tlv[LLDP.Neighbors.TLV_SUBTYPE] == 3,
                test_neighbor,
            )
        )
        assert len(tlvs) == 1
        test_vlan = tlvs[0][VLANS_SUBTREE][0]
        assert test_vlan[NAME] == "v2-0488-03-0505"


def test_lldp_mfs(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        test_neighbor = lldp_config[LLDP.NEIGHBORS_SUBTREE][0]

        tlvs = list(
            filter(
                lambda tlv: tlv[LLDP.Neighbors.TLV_TYPE] == 127
                and tlv[LLDP.Neighbors.ORGANIZATION_CODE] == "00:12:0f"
                and tlv[LLDP.Neighbors.TLV_SUBTYPE] == 4,
                test_neighbor,
            )
        )
        assert len(tlvs) == 1
        assert tlvs[0][MFS_KEY] == 1522


def test_lldp_empty_neighbors(lldptest_up):
    with lldp_enabled(lldptest_up):
        dstate = statelib.show_only((LLDPTEST,))
        lldp_state = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert not lldp_state.get(LLDP.NEIGHBORS_SUBTREE, [])


def test_show_running_config_has_no_lldp_neighbor(lldptest_up):
    with lldp_enabled(lldptest_up):
        _send_lldp_packet()
        dstate = statelib.show_only((LLDPTEST,))
        lldp_config = dstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
        assert len(lldp_config[LLDP.NEIGHBORS_SUBTREE]) == 1
        running_config = libnmstate.show_running_config()
        for iface_config in running_config[Interface.KEY]:
            if iface_config[Interface.NAME] == LLDPTEST:
                lldp_iface_config = iface_config
                break
        assert lldp_iface_config[LLDP.CONFIG_SUBTREE][LLDP.ENABLED]
        assert (
            LLDP.NEIGHBORS_SUBTREE
            not in lldp_iface_config[LLDP.CONFIG_SUBTREE]
        )


@contextmanager
def lldp_enabled(ifstate):
    lldp_config = ifstate[Interface.KEY][0][LLDP.CONFIG_SUBTREE]
    lldp_config[LLDP.ENABLED] = True
    libnmstate.apply(ifstate)
    try:
        yield ifstate
    finally:
        lldp_config[LLDP.ENABLED] = False
        libnmstate.apply(ifstate)


def _send_lldp_packet():
    test_dir = os.path.dirname(os.path.realpath(__file__))
    cmdlib.exec_cmd(
        f"ip netns exec {LLDP_TEST_NS} "
        f"tcpreplay --intf1={LLDPTEST_PEER} "
        f"{test_dir}/test_captures/lldp.pcap".split(),
        check=True,
    )
    time.sleep(1)


@pytest.fixture
def eth1_up_with_lldp(eth1_up):
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    LLDP.CONFIG_SUBTREE: {LLDP.ENABLED: True},
                }
            ]
        }
    )


def test_show_saved_config_ethernet_down_with_lldp(eth1_up_with_lldp):
    running_state = statelib.show_only(("eth1",))
    libnmstate.apply(
        {
            Interface.KEY: [
                {
                    Interface.NAME: "eth1",
                    Interface.STATE: InterfaceState.DOWN,
                }
            ]
        }
    )
    saved_state = statelib.show_saved_config_only(("eth1",))

    assert saved_state[Interface.KEY][0][LLDP.CONFIG_SUBTREE] == {
        LLDP.ENABLED: True
    }
    assertlib.assert_state_match_full(saved_state, running_state)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
import os
import time

import pytest

import libnmstate

from .testlib import ifacelib


def get_current_open_fd():
    time.sleep(0.1)  # Wait sysfs/proc been updated.
    return len(os.listdir("/proc/self/fd"))


@pytest.fixture(scope="function")
def disable_logging():
    logger = logging.getLogger()
    logger.disabled = True
    try:
        yield
    finally:
        logger.disabled = False


@pytest.mark.tier1
def test_libnmstate_show_fd_leak(disable_logging):
    original_fd = get_current_open_fd()
    for x in range(0, 100):
        libnmstate.show()
    assert get_current_open_fd() <= original_fd


@pytest.mark.tier1
def test_libnmstate_apply_fd_leak(disable_logging):
    original_fd = get_current_open_fd()
    for x in range(0, 10):
        with ifacelib.iface_up("eth1"):
            pass
    assert get_current_open_fd() <= original_fd
